import numpy as np
import vegas as vg

from DarkNews import logger, prettyprinter

from DarkNews import const
from DarkNews import pdg
from DarkNews import integrands
from DarkNews import MC
from DarkNews import model
from DarkNews import amplitudes as amps
from DarkNews import phase_space as ps


class UpscatteringProcess:
    """ 
        Describes the process of upscattering with arbitrary vertices and masses
    
    """

    def __init__(self, nu_projectile, nu_upscattered, nuclear_target, scattering_regime, TheoryModel, helicity):

        self.nuclear_target = nuclear_target
        self.scattering_regime = scattering_regime
        if self.scattering_regime == "coherent":
            self.target = self.nuclear_target
        elif self.scattering_regime == "p-el":
            self.target = self.nuclear_target.get_constituent_nucleon("proton")
        elif self.scattering_regime == "n-el":
            self.target = self.nuclear_target.get_constituent_nucleon("neutron")
        elif self.scattering_regime == "DIS":
            self.target = self.nuclear_target.get_constituent_quarks()
        else:
            logger.error(f"Scattering regime {scattering_regime} not supported.")

        # How many constituent targets inside scattering regime?
        if self.scattering_regime == "coherent":
            self.target_multiplicity = 1
        elif self.scattering_regime == "p-el":
            self.target_multiplicity = self.nuclear_target.Z
        elif self.scattering_regime == "n-el":
            self.target_multiplicity = self.nuclear_target.N
        else:
            logger.error(f"Scattering regime {self.scattering_regime} not supported.")

        self.nu_projectile = nu_projectile
        self.nu_upscattered = nu_upscattered
        self.TheoryModel = TheoryModel
        self.helicity = helicity

        self.MA = self.target.mass
        self.mzprime = TheoryModel.mzprime if TheoryModel.mzprime is not None else 1e10
        self.mhprime = TheoryModel.mhprime if TheoryModel.mhprime is not None else 1e10
        self.m_ups = self.nu_upscattered.mass

        if self.helicity == "conserving":
            self.h_upscattered = -1
        elif self.helicity == "flipping":
            self.h_upscattered = +1
        else:
            logger.error(f"Error! Could not find helicity case {self.helicity}")

        self.Cij = TheoryModel.c_aj[pdg.get_lepton_index(nu_projectile), pdg.get_HNL_index(nu_upscattered)]
        self.Cji = self.Cij
        self.Vij = TheoryModel.d_aj[pdg.get_lepton_index(nu_projectile), pdg.get_HNL_index(nu_upscattered)]
        self.Vji = self.Vij
        self.Sij = TheoryModel.s_aj[pdg.get_lepton_index(nu_projectile), pdg.get_HNL_index(nu_upscattered)]
        self.Sji = self.Sij
        self.Tij = TheoryModel.t_aj[pdg.get_lepton_index(nu_projectile), pdg.get_HNL_index(nu_upscattered)]
        self.Tji = self.Tij

        ###############
        # Hadronic vertices
        if self.target.is_nucleus:
            # self.Chad = const.gweak/4.0/const.cw*np.abs((1.0-4.0*const.s2w)*self.target.Z-self.target.N)
            # self.Vhad = const.eQED*TheoryModel.epsilon*self.target.Z
            self.Chad = TheoryModel.cprotonV * self.target.Z + TheoryModel.cneutronV * self.target.N
            self.Vhad = TheoryModel.dprotonV * self.target.Z + TheoryModel.dneutronV * self.target.N
            self.Shad = TheoryModel.dprotonS * self.target.Z + TheoryModel.dneutronS * self.target.N
        elif self.target.is_proton:
            self.Chad = TheoryModel.cprotonV
            self.Vhad = TheoryModel.dprotonV
            self.Shad = TheoryModel.dprotonS
        elif self.target.is_neutron:
            self.Chad = TheoryModel.cneutronV
            self.Vhad = TheoryModel.dneutronV
            self.Shad = TheoryModel.dneutronS

        # If three portal model, set mass-mixed vertex
        if isinstance(TheoryModel, model.ThreePortalModel):
            # mass mixed vertex
            self.Cprimehad = self.Chad * TheoryModel.epsilonZ
        else:
            self.Cprimehad = 0.0

        # Neutrino energy threshold
        self.Ethreshold = self.m_ups ** 2 / 2.0 / self.MA + self.m_ups

        # vectorize total cross section calculator using vegas integration
        self.vectorized_total_xsec = np.vectorize(self.scalar_total_xsec, excluded=["self", "diagram", "NINT", "NEVAL", "NINT_warmup", "NEVAL_warmup"])

        self.calculable_diagrams = find_calculable_diagrams(TheoryModel)

    def scalar_total_xsec(self, Enu, diagram="total", NINT=MC.NINT, NEVAL=MC.NEVAL, NINT_warmup=MC.NINT_warmup, NEVAL_warmup=MC.NEVAL_warmup):
        # below threshold
        if Enu < (self.Ethreshold):
            return 0.0
        else:
            DIM = 1
            batch_f = integrands.UpscatteringXsec(dim=DIM, Enu=Enu, ups_case=self, diagram=diagram)
            integ = vg.Integrator(DIM * [[0.0, 1.0]])  # unit hypercube

            integrals = MC.run_vegas(batch_f, integ, adapt_to_errors=True, NINT=NINT, NEVAL=NEVAL, NINT_warmup=NINT_warmup, NEVAL_warmup=NEVAL_warmup)
            logger.debug(f"Main VEGAS run completed.")

            return integrals["diff_xsec"].mean * batch_f.norm["diff_xsec"]

    def total_xsec(self, Enu, diagrams=["total"], NINT=MC.NINT, NEVAL=MC.NEVAL, NINT_warmup=MC.NINT_warmup, NEVAL_warmup=MC.NEVAL_warmup):
        """ 
            Returns the total upscattering xsec for a fixed neutrino energy in cm^2
        """
        self.Enu = Enu
        all_xsecs = 0.0
        for diagram in diagrams:
            if diagram in self.calculable_diagrams or diagram == "total":
                tot_xsec = self.vectorized_total_xsec(Enu, diagram=diagram, NINT=NINT, NEVAL=NEVAL, NINT_warmup=NINT_warmup, NEVAL_warmup=NEVAL_warmup)
            else:
                logger.warning(f"Warning: Diagram not found. Either not implemented or misspelled. Setting tot xsec it to zero: {diagram}")
                tot_xsec = 0.0 * Enu

            #############
            # integrated xsec coverted to cm^2
            all_xsecs += tot_xsec * const.attobarn_to_cm2 * self.target_multiplicity
            logger.debug(f"Total cross section for {diagram} calculated.")

        return all_xsecs

    def diff_xsec_Q2(self, Enu, Q2, diagrams=["total"]):
        """ 
            Returns the differential upscattering xsec for a fixed neutrino energy in cm^2
        """
        s = Enu * self.MA * 2 + self.MA ** 2
        physical = (Q2 > ps.upscattering_Q2min(Enu, self.m_ups, self.MA)) & (Q2 < ps.upscattering_Q2max(Enu, self.m_ups, self.MA))
        diff_xsecs = amps.upscattering_dxsec_dQ2([s, -Q2, 0.0], process=self, diagrams=diagrams)
        if type(diff_xsecs) is dict:
            return {key: diff_xsecs[key] * physical for key in diff_xsecs.keys()}
        else:
            return diff_xsecs * physical * const.attobarn_to_cm2 * self.target_multiplicity


class FermionDileptonDecay:
    def __init__(self, nu_parent, nu_daughter, final_lepton1, final_lepton2, TheoryModel, h_parent=-1):

        self.TheoryModel = TheoryModel
        self.HNLtype = TheoryModel.HNLtype
        self.h_parent = h_parent

        # particle masses
        self.mzprime = TheoryModel.mzprime if TheoryModel.mzprime is not None else 1e10
        self.mhprime = TheoryModel.mhprime if TheoryModel.mhprime is not None else 1e10
        self.mm = final_lepton1.mass * const.MeV_to_GeV
        self.mp = final_lepton2.mass * const.MeV_to_GeV

        # Neutral lepton vertices
        if nu_daughter == pdg.nulight:
            self.Cih = np.sqrt(np.sum(np.abs(TheoryModel.c_aj[const.inds_active, pdg.get_HNL_index(nu_parent)]) ** 2))
            self.Dih = np.sqrt(np.sum(np.abs(TheoryModel.d_aj[const.inds_active, pdg.get_HNL_index(nu_parent)]) ** 2))
            self.Sih = np.sqrt(np.sum(np.abs(TheoryModel.s_aj[const.inds_active, pdg.get_HNL_index(nu_parent)]) ** 2))
            self.Tih = np.sqrt(np.sum(np.abs(TheoryModel.t_aj[const.inds_active, pdg.get_HNL_index(nu_parent)]) ** 2))
        else:
            self.Cih = TheoryModel.c_aj[pdg.get_HNL_index(nu_daughter), pdg.get_HNL_index(nu_parent)]
            self.Dih = TheoryModel.d_aj[pdg.get_HNL_index(nu_daughter), pdg.get_HNL_index(nu_parent)]
            self.Sih = TheoryModel.s_aj[pdg.get_HNL_index(nu_daughter), pdg.get_HNL_index(nu_parent)]
            self.Tih = TheoryModel.t_aj[pdg.get_HNL_index(nu_daughter), pdg.get_HNL_index(nu_parent)]

        # Charged lepton vertices
        self.Cv = TheoryModel.ceV
        self.Ca = TheoryModel.ceA
        self.Dv = TheoryModel.deV
        self.Da = TheoryModel.deA
        self.Ds = TheoryModel.deS
        self.Dp = TheoryModel.deP

        if nu_parent == pdg.neutrino4:
            self.m_parent = TheoryModel.m4
        elif nu_parent == pdg.neutrino5:
            self.m_parent = TheoryModel.m5
        elif nu_parent == pdg.neutrino6:
            self.m_parent = TheoryModel.m6
        else:
            self.m_parent = 0.0

        if nu_daughter == pdg.neutrino4:
            self.m_daughter = TheoryModel.m4
        elif nu_daughter == pdg.neutrino5:
            self.m_daughter = TheoryModel.m5
        elif nu_daughter == pdg.neutrino6:
            self.m_daughter = TheoryModel.m6
        else:
            self.m_daughter = 0.0

        # check if CC is allowed
        # CC_mixing1 = LNC, CC_mixing2 = LNV channel.
        if pdg.in_same_doublet(nu_daughter, final_lepton1):
            self.CC_mixing1 = TheoryModel.Ulep[pdg.get_lepton_index(final_lepton1), pdg.get_HNL_index(nu_parent)]
            self.CC_mixing2 = TheoryModel.Ulep[pdg.get_lepton_index(final_lepton2), pdg.get_HNL_index(nu_parent)]
        else:
            self.CC_mixing1 = 0
            self.CC_mixing2 = 0
        ## Minus sign important for interference!
        self.CC_mixing2 *= -1

        if self.m_parent - self.m_daughter - self.mm - self.mp < 0:
            logger.error(f"Error! Final states have an excess in mass of {self.m_parent - self.m_daughter - self.mm - self.mp} on top of parent particle.")
            raise ValueError("Energy not conserved.")

        ## Is the mediator on shell?
        self.on_shell = (self.m_parent - self.m_daughter > TheoryModel.mzprime) and (TheoryModel.mzprime > self.mm + self.mp)
        self.off_shell = not self.on_shell
        ## does it have transition magnetic moment?
        self.TMM = TheoryModel.has_TMM


class FermionSinglePhotonDecay:
    def __init__(self, nu_parent, nu_daughter, TheoryModel, h_parent=-1):

        self.TheoryModel = TheoryModel
        self.HNLtype = TheoryModel.HNLtype
        self.h_parent = h_parent

        # mass of the HNLs
        if nu_daughter == pdg.neutrino4:
            self.m_daughter = TheoryModel.m4
        elif nu_daughter == pdg.neutrino5:
            self.m_daughter = TheoryModel.m5
        elif nu_daughter == pdg.neutrino6:
            self.m_daughter = TheoryModel.m6
        else:
            self.m_daughter = 0.0

        if nu_parent == pdg.neutrino4:
            self.m_parent = TheoryModel.m4
        elif nu_parent == pdg.neutrino5:
            self.m_parent = TheoryModel.m5
        elif nu_parent == pdg.neutrino6:
            self.m_parent = TheoryModel.m6
        else:
            self.m_parent = 0.0

        # transition magnetic moment parameter
        if nu_daughter == pdg.nulight:
            # |T| = sqrt(|T_ei|^2 + |T_mui|^2 + |T_taui|^2)
            self.Tih = np.sqrt(np.sum(np.abs(self.TheoryModel.t_aj[const.inds_active, pdg.get_HNL_index(nu_parent)]) ** 2))
            self.m_daughter = 0.0
        else:
            self.Tih = self.TheoryModel.t_aj[pdg.get_HNL_index(nu_daughter), pdg.get_HNL_index(nu_parent)]


def find_calculable_diagrams(bsm_model):
    """ 
    Args:
        bsm_model (DarkNews.model.Model): main BSM model class of DarkNews

    Returns:
        list: with all non-zero upscattering diagrams to be computed in this model.
    """

    calculable_diagrams = []
    calculable_diagrams.append("NC_SQR")
    if bsm_model.has_vector_coupling:
        calculable_diagrams.append("KinMix_SQR")
        calculable_diagrams.append("KinMix_NC_inter")
    if bsm_model.is_mass_mixed:
        calculable_diagrams.append("MassMix_SQR")
        calculable_diagrams.append("MassMix_NC_inter")
        if bsm_model.has_vector_coupling:
            calculable_diagrams.append("KinMix_MassMix_inter")
    if bsm_model.has_TMM:
        calculable_diagrams.append("TMM_SQR")
    if bsm_model.has_scalar_coupling:
        calculable_diagrams.append("Scalar_SQR")
        calculable_diagrams.append("Scalar_NC_inter")
        if bsm_model.has_vector_coupling:
            calculable_diagrams.append("Scalar_KinMix_inter")
        if bsm_model.is_mass_mixed:
            calculable_diagrams.append("Scalar_MassMix_inter")
    return calculable_diagrams



# This module is a customisation of fourFn.py from PyParsing module examples
# https://github.com/pyparsing/pyparsing/blob/203fa36d7ae6b79344e4bf13531b77c09f313793/examples/fourFn.py

# fourFn.py
#
# Demonstration of the pyparsing module, implementing a simple 4-function expression parser,
# with support for scientific notation, and symbols for e and pi.
# Extended to add exponentiation and simple built-in functions.
# Extended test cases, simplified pushFirst method.
# Removed unnecessary expr.suppress() call (thanks Nathaniel Peterson!), and added Group
# Changed fnumber to use a Regex, which is now the preferred method
# Reformatted to latest pypyparsing features, support multiple and variable args to functions
#
# Copyright 2003-2019 by Paul McGuire
#
from xmlrpc.client import Boolean
from pyparsing import (
    Literal,
    Word,
    Group,
    Forward,
    alphas,
    alphanums,
    Regex,
    ParseException,
    CaselessKeyword,
    Suppress,
    delimitedList,
    Or,
    QuotedString,
)
import math
import operator


class AssignmentParser:
    """
        Parse an expression of the kind
        var_z = 5 + exp(4*sin(PI/3)) + var_a
        where it does both computation of the expression without calling eval,
        and substitution with other variables (var_a), previously defined.
        It stores the variables in a dictionary, which is passed by argument.
        It is the same dictionary on which it does the lookup.
    """

    # map operator symbols to corresponding arithmetic operations
    epsilon = 1e-100
    opn = {
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,
        "/": operator.truediv,
        "^": operator.pow,
    }

    # functions
    fn = {
        "sin": math.sin,
        "cos": math.cos,
        "tan": math.tan,
        "exp": math.exp,
        "abs": abs,
        "trunc": int,
        "round": round,
        "sgn": lambda a: -1 if a < -AssignmentParser.epsilon else 1 if a > AssignmentParser.epsilon else 0,
        "list": lambda *args: list([*args]),
        "sum": sum,
        "str": str,
    }

    class ParsingError(Exception):
        """ Deal with errors coming only from this parser """

        pass

    def __init__(self, parameters):
        self.parameters = parameters
        self._bnf = None
        self._current_stack = []
        # load the grammar
        self._BNF()

    def _BNF(self):
        """
            It means Backus-Naur Form: this is the main definition of the grammar
            expop    :: '^'
            multop   :: '*' | '/'
            addop    :: '+' | '-'
            equalop  :: '='
            integer  :: ['+' | '-'] '0'..'9'+
            ident    :: [ a-zA-Z_ ][ a-zA-Z0-9_ ]*
            atom     :: PI | E | real | fn '(' expr ')' | '(' expr ')'
            factor   :: atom [ expop factor ]*
            term     :: factor [ multop factor ]*
            expr     :: variable equalop term [ addop term ]*
        """
        if not self._bnf:
            # use CaselessKeyword for e and pi, to avoid accidentally matching
            # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
            # and CaselessKeyword only match whole words
            e = CaselessKeyword("E")
            pi = CaselessKeyword("PI")
            # fnumber = Combine(Word("+-"+nums, nums) +
            #                    Optional("." + Optional(Word(nums))) +
            #                    Optional(e + Word("+-"+nums, nums)))
            # or use provided pyparsing_common.number, but convert back to str:
            # fnumber = ppc.number().addParseAction(lambda t: str(t[0]))
            fnumber = Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
            ident = Word(alphas, alphanums + "_")  # variables can have names starting only with letters, no underscore nor numbers

            plus, minus, mult, div = map(Literal, "+-*/")
            lpar, rpar = map(Suppress, "()")
            l_delim, r_delim = map(Suppress, "[]")
            addop = plus | minus
            multop = mult | div
            expop = Literal("^")
            equalop = Literal("=")

            expr = Forward()
            expr_list = delimitedList(Group(expr))

            # Quoted strings
            quoted_str_class = QuotedString(quoteChar='"', escChar="\\", multiline=True, unquoteResults=False) | QuotedString(  # double quote string
                quoteChar="'", escChar="\\", multiline=True, unquoteResults=False
            )  # single quote string  # keep the quotes and use them to separate strings from variables
            quoted_str = quoted_str_class.setParseAction(self._push_first)

            # Lists
            list_elements = delimitedList(Group(quoted_str | expr))
            list_pattern = l_delim - Group(list_elements) + r_delim

            # Function call: need to encapsulate again list_pattern in a group otherwise sum([list]) would detect one more argument, but it is only one: that's because the list is converted to [('list', n), [...n elements...]], so 2 elements, but it is only one if it is a parameter of a function call
            fn_call = (ident + lpar - Group(expr_list | Group(list_pattern)) + rpar).setParseAction(self._insert_fn_argcount_tuple)
            atom = (addop[...] + ((fn_call | pi | e | fnumber | ident).setParseAction(self._push_first) | Group(lpar + expr + rpar))).setParseAction(
                self._push_unary_minus
            )

            # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left
            # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
            factor = Forward()
            factor <<= atom + (expop + factor).setParseAction(self._push_first)[...]
            term = factor + (multop + factor).setParseAction(self._push_first)[...]
            expr <<= term + (addop + term).setParseAction(self._push_first)[...]
            assignment_expr = Forward()
            # _push_first should be given only to the full match otherwise if given when it matches singularly then it would repeat everything every time it backtracks
            assignment_expr <<= (
                ident
                + equalop
                + Or(
                    [
                        quoted_str,
                        list_pattern.setParseAction(self._insert_list_argcount_tuple, self._push_first),
                        expr,  # this is the most-specific patterns should be placed ahead
                    ]
                )
            ).setParseAction(self._push_first)

            # assignment_expr <<= (ident + equalop + list_pattern.setParseAction(self._insert_list_argcount_tuple, self._push_first)).setParseAction(self._push_first)
            self._bnf = assignment_expr
        return self._bnf

    def _insert_fn_argcount_tuple(self, t):
        """ Parse action that replaces the function identifier with a (name, number of args, "function") tuple. """
        fn = t.pop(0)
        num_args = len(t[0])
        t.insert(0, (fn, num_args))

    def _insert_list_argcount_tuple(self, t):
        """ Parse action that replaces the list found with a ("list", number of elements) tuple. """
        num_args = len(t[0])
        t.insert(0, ("list", num_args))

    def _push_first(self, toks):
        self._current_stack.append(toks[0])

    def _push_unary_minus(self, toks):
        for t in toks:
            if t == "-":
                self._current_stack.append("unary -")
            else:
                break

    def _evaluate_expression(self, s):
        op, num_args = s.pop(), 0
        if isinstance(op, tuple):
            op, num_args = op
        if isinstance(op, bool):
            return op
        if op == "unary -":
            return -self._evaluate_expression(s)
        if op in "+-*/^":
            # note: operands are pushed onto the stack in reverse order
            op2 = self._evaluate_expression(s)
            op1 = self._evaluate_expression(s)
            return self.opn[op](op1, op2)
        elif op == "PI":
            return math.pi  # 3.1415926535
        elif op == "E":
            return math.e  # 2.718281828
        elif op.lower() == "true":
            return True
        elif op.lower() == "false":
            return False
        elif op == "str":  # so far str will consume only one argument
            return str(s.pop())
        elif op in self.fn.keys():
            # note: args are pushed onto the stack in reverse order
            args = reversed([self._evaluate_expression(s) for _ in range(num_args)])
            return self.fn[op](*args)
        elif op[0].isalpha():
            try:
                return self.parameters[op]
            except KeyError:
                raise self.ParsingError(f"Invalid identifier '{op}'")
        else:
            # try to evaluate as float, then as string
            try:
                return float(op)
            except ValueError:
                return str(op.strip('"').strip("'"))

    def evaluate_stack(self, copy=False):
        # if copy == True, then don't consume the stack
        # get the first element of the stack, which must be the variable name
        if copy:
            stack = self._current_stack[:]
        else:
            stack = self._current_stack
        var_name = stack.pop()
        # consistency check: should not be equal to E, e, PI, pi, Pi, or any other function name
        if (var_name.lower() in ["e", "pi", "true", "false"]) or (var_name in self.fn.keys()):
            raise self.ParsingError(f"Variable name '{var_name}' not allowed.")
        # evaluate the rest of the stack and assign the result to the variable
        self.parameters[var_name] = self._evaluate_expression(
            stack
        )  # if evaluation contains the new variable which we are trying to assign, everything will fail by construction
        return var_name, self.parameters[var_name]

    def clean_stack(self):
        self._current_stack[:] = []

    def parse_string(self, *args, **kwargs):
        self.clean_stack()
        return self._bnf.parseString(*args, **kwargs)

    def scan_string(self, *args, **kwargs):
        self.clean_stack()
        return self._bnf.scanString(*args, **kwargs)

    def parse_file(self, file, comments="#"):
        # read file
        try:
            # assume it is an open file
            lines = file.readlines()
        except AttributeError:
            # assume it is a str to file path
            with open(file, "r", encoding="utf8") as f:
                lines = f.readlines()

        # create a clean text without blank lines and comments
        clean_text = ""
        for i, line in enumerate(lines):
            partition = line.partition(comments)[0]
            if partition.strip() == "":
                continue
            clean_text += partition + "\n"
        for tokens, i_beg, i_end in self.scan_string(clean_text):
            # the effective parsing is done when the generator is called
            # so the stack fills at each iteration
            try:
                self.evaluate_stack()
            except ParseException as pe:
                print(partition, f"Failed parse (start: {i_beg}, end: {i_end}):", str(pe))
            except self.ParsingError as e:
                print(partition, f"Failed evaluation (start: {i_beg}, end: {i_end}):", str(e))
            finally:
                self.clean_stack()  # clean the stack in any case, because if there are errors, then we need to have a clean list before the next iteration


if __name__ == "__main__":
    parser = AssignmentParser({})

    def test(s, expected_name, expected_value, should_fail=False):
        should_fail_string = "[\033[92mOK\033[0m]" if should_fail else "[\033[91mNO\033[0m]"
        try:
            results = parser.parse_string(s, parseAll=True)
            name, val = parser.evaluate_stack(copy=True)
        except ParseException as pe:
            print(should_fail_string, s, "failed parse:", str(pe))
        except AssignmentParser.ParsingError as e:
            print(should_fail_string, s, "failed eval:", str(e), parser._current_stack)
        else:
            condition_value = (val == expected_value) if not isinstance(expected_value, list) else all([v == ev for v, ev in zip(val, expected_value)])
            if (name == expected_name) and condition_value:
                print("[\033[92mOK\033[0m]", s, "<=>", name, "=", val, results, "=>", parser._current_stack)
            else:
                print("[\033[91mNO\033[0m]", s, "<=>", name, "=", val, "!!!", expected_name, "=", expected_value, results, "=>", parser._current_stack)

    class TEST:
        gD = 9.7543
        A = gD * 9.7543
        B = 9.7543 ** 2 + gD
        C = 9.7543 + 2 - 4 * A
        D = 9.7543 + math.sin(0.3)
        E = D
        alphaD = gD ** 2 / (4 * math.pi)
        sinx = math.sin(math.pi / 3) - 8
        hbar = 6.582119569e-25
        c = 299792458.0
        a_variable = c ** 2 * 3.2e-4 / math.sin(math.pi / 7) + 12 * math.exp(-2 * abs(hbar))

    test("gD = 9.7543", "gD", TEST.gD)
    test("A = gD * 9.7543", "A", TEST.A)
    test("B = 9.7543^2 + gD", "B", TEST.B)
    test("C = 9.7543 + 2 - 4*A", "C", TEST.C)
    test("D = 9.7543 + sin(0.3)", "D", TEST.D)
    test("E = 9.7543 + sin(0.3)", "E", TEST.E, should_fail=True)
    test("alphaD = gD^2 / (4 * PI)", "alphaD", TEST.alphaD)
    test("sinx = sin(PI/3) - 8.", "sinx", TEST.sinx)
    test("number = sinx * 5.35e6", "number", TEST.sinx * 5.35e6)
    test("number2 = gD^2 * 5.35e6", "number2", TEST.gD ** 2 * 5.35e6)
    test("exp = exp(3*PI)", "exp", math.exp(3 * math.pi), should_fail=True)
    test("exp_0 = exp(3*PI)", "exp_0", math.exp(3 * math.pi))
    test("24ff = -10+tan(PI/4)^2", "24ff", -10 + math.tan(math.pi / 4) ** 2, should_fail=True)
    test("ff24 = -10+tan(PI/4)^2", "ff24", -10 + math.tan(math.pi / 4) ** 2)
    test(" ", "", 0, should_fail=True)
    test("hbar = 6.582119569e-25", "hbar", TEST.hbar)
    test("c = 299792458.0", "c", TEST.c)
    test("a_variable = c^2 * 3.2e-4 / sin(PI/7) + 12 * exp( -2 * abs(hbar) )", "a_variable", TEST.a_variable)
    test(
        """multi_line = c^2 * 3.2e-4 / sin(PI/7) + 
    12 * exp( -2 * abs(hbar) )""",
        "multi_line",
        TEST.a_variable,
    )
    test('s_1 = "hello world"', "s_1", "hello world")
    test('s_2 = "hello world" "people"', "s_2", None, should_fail=True)
    test("s_3 = 'hello world'", "s_3", "hello world")
    test("test_1 = true", "test_1", True)
    test("test_2 = False", "test_2", False)
    test("test_3 = True False", "test_3", None, should_fail=True)
    test('test_4 = True "hello"', "test_4", None, should_fail=True)
    test('test_5 = exp(3*PI)*c True "hello"', "test_5", None, should_fail=True)
    test('test_6 = exp(3*PI)*c + True - "hello"', "test_6", None, should_fail=True)
    test("test_7 = exp(3*PI)*c + True", "test_7", math.exp(3 * math.pi) * TEST.c + True)  # it's ok, because float + bool is the bool converted to integer
    test("test_8 = False + exp(3*PI)*c", "test_8", False + math.exp(3 * math.pi) * TEST.c)
    test("list_1 = [-10, 2.3]", "list_1", [-10, 2.3])
    test("sum_1 = sum(list_1)", "sum_1", sum([-10, 2.3]))
    test("sum_2 = sum([-10+tan(PI/4)^2, exp(3*PI)*E])", "sum_2", sum([-10 + math.tan(math.pi / 4) ** 2, math.exp(3 * math.pi) * math.e]))
    test("list_2 = [-10+tan(PI/4)^2, exp(3*PI)*E]", "list_2", [-10 + math.tan(math.pi / 4) ** 2, math.exp(3 * math.pi) * math.e])
    test('list_3 = ["hello", exp(3*PI)*E]', "list_3", ["hello", math.exp(3 * math.pi) * math.e])
    test('list_4 = ["hello", "world"]', "list_4", ["hello", "world"])
    test('list_5 = ["hello world" "people", 10]', "list_5", None, should_fail=True)
    test('list_6 = ["hello", a_variable, "world"]', "list_6", ["hello", TEST.a_variable, "world"])
    test(
        """list_7 = [
                \"hello\",
                a_variable,
                \"world\",
                \"Lorem ipsum dolor sit amet, consectetur adipiscing elit.
Nunc ullamcorper blandit nibh, vitae congue lacus convallis at.
Donec interdum, ex et fermentum aliquam, ante magna convallis magna, et molestie quam quam et dolor.\",
                c^2 * 3.2e-4 / sin(PI/7) + 
    12 * exp( -2 * abs(hbar) )
            ]""",
        "list_7",
        [
            "hello",
            TEST.a_variable,
            "world",
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit.\nNunc ullamcorper blandit nibh, vitae congue lacus convallis at.\nDonec interdum, ex et fermentum aliquam, ante magna convallis magna, et molestie quam quam et dolor.",
            TEST.a_variable,
        ],
    )
    test(
        """nuclear_targets = [
    'H1',
    'C12',
    'O16',
    'Cu63',
    "Zn64",
    'Pb208'
]""",
        "nuclear_targets",
        ["H1", "C12", "O16", "Cu63", "Zn64", "Pb208"],
    )
    test(
        """fiducial_mass_per_target = [
    0.42, 
    5.4, 
    3.32, 
    0.4, 
    0.8, 
    12.0
] """,
        "fiducial_mass_per_target",
        [0.42, 5.4, 3.32, 0.4, 0.8, 12.0],
    )
    test("fiducial_mass = sum(fiducial_mass_per_target)", "fiducial_mass", sum([0.42, 5.4, 3.32, 0.4, 0.8, 12.0]))

    # print stored variables
    print("\nStored variables")
    for k, v in parser.parameters.items():
        print(k, "=", v, type(v))

    # test multi assignments
    parser_multi = AssignmentParser({})
    text = """
hbar = 6.582119569e-25
c = 299792458.0

a_variable = -10+tan(PI/4)^2
a_list = [
                \"hello\",
                a_variable,
                \"world\",
                \"Lorem ipsum dolor sit amet, consectetur adipiscing elit.
Nunc ullamcorper blandit nibh, vitae congue lacus convallis at.
Donec interdum, ex et fermentum aliquam, ante magna convallis magna, et molestie quam quam et dolor.\",
                c^2 * 3.2e-4 / sin(PI/7) + 
    12 * exp( -2 * abs(hbar) )
            ]


sinx = sin(PI/3) - 8.

a_multi_line_expression = c^2 * 3.2e-4 / sin(PI/7) + 
    12 * exp( -2 * abs(hbar) )
a_bool = True

a_string = \"hello world\"
    """

    matches = parser_multi.scan_string(text)
    for _ in matches:
        try:
            print(parser_multi._current_stack)
            name, val = parser_multi.evaluate_stack(copy=True)
        except ParseException as pe:
            print("failed parse:", str(pe))
        except AssignmentParser.ParsingError as e:
            print("failed eval:", str(e), parser_multi._current_stack)
        finally:
            parser_multi._current_stack[:] = []

    print("\nStored variables")
    for k, v in parser_multi.parameters.items():
        print(k, "=", v, type(v))



import numpy as np

from DarkNews import Cfourvec as Cfv
from DarkNews import logger

X = 0
Y = 1
Z = 2


##############################
# numpy functions
def dot4(x, y):
    if np.size(x) != 4 or np.size(y) != 4:
        logger.error(f"Error! The dot4 product of two vectors with sizes different from 4 is not defined. Sizes s(x)={np.shape(x)} and s(y)={np.shape(y)}.")
        return 0
    return x[0] * y[0] - x[1] * y[1] - x[2] * y[2] - x[3] * y[3]


def dot3(x, y):
    if np.size(x) != 4 or np.size(y) != 4:
        logger.error(f"Error! The dot3 product of two vectors with sizes different from 4 is not defined. Sizes s(x)={np.shape(x)} and s(y)={np.shape(y)}.")
        return x[1] * y[1] + x[2] * y[2] + x[3] * y[3]


def cos_opening_angle(x, y):
    return Cfv.get_cos_opening_angle(x, y)


def cos_azimuthal(x):
    return Cfv.get_cosTheta(x)


def inv_mass(x, y):
    return Cfv.inv_mass(x, y)


# THRESHOLD = 0.0
# def inv_mass(x,y):
# 	mSQR = np.clip(dot4(x,y), THRESHOLD, np.inf)
# 	if (mSQR < 0).any():
# 		logger.warning("Warning! Trying to compute invariant mass with negative (p_a.p_b) product. Possibly a numerical instability?")
# 	return np.sqrt(mSQR)


def mass(x):
    return inv_mass(x, x)


def get_3vec(x):
    return x[1:]


def get_direction(x):
    if np.size(x) != 4:
        print("ERROR! Wrong size np.shape(x):", np.shape(x))
        return 0
    return get_3vec(x) / np.sqrt(dot3(x, x))


##############################
# dataframe functions
def df_dot4(dx, dy):
    return Cfv.dot4(dx.to_numpy(), dy.to_numpy())


def df_dot3(dx, dy):
    return Cfv.dot3(dx.to_numpy(), dy.to_numpy())


def df_cos_opening_angle(dx, dy):
    return Cfv.get_cos_opening_angle(dx.to_numpy(), dy.to_numpy())


def df_cos_azimuthal(dx):
    return cos_azimuthal(dx.to_numpy())


def df_inv_mass(dx, dy):
    return inv_mass(dx.to_numpy(), dy.to_numpy())


##############################
# transformation functions
def R(v4, theta, i):
    c, s = np.cos(theta), np.sin(theta)
    if i == X:
        R = np.array(((1.0, 0, 0, 0), (0, 1.0, 0, 0), (0, 0, c, -s), (0, 0, s, c)))
    if i == Y:
        R = np.array(((1.0, 0, 0, 0), (0, c, 0, -s), (0, 0, 1.0, 0), (0, s, 0, c)))
    if i == Z:
        R = np.array(((1.0, 0, 0, 0), (0, c, -s, 0), (0, s, c, 0), (0, 0, 0, 1.0)))
    return R.dot(v4)


def L(v4, beta):
    gamma = 1.0 / np.sqrt(1.0 - beta * beta)
    R = np.array(((gamma, 0, 0, -gamma * beta), (0, 1, 0, 0), (0, 0, 1, 0), (-gamma * beta, 0, 0, gamma)))
    return R.dot(v4)


def T(v4, beta, theta, phi):
    return L(R(R(v4, -phi, Z), theta, Y), -beta)


def Tinv(v4, beta, theta, phi):
    return R(R(L(v4, beta), -theta, Y), phi, Z)



#!/usr/bin/env python3

import argparse
from DarkNews.GenLauncher import GenLauncher
import DarkNews.parsing_tools as pt
import os
import subprocess
import shutil
from pathlib import Path

def dn_gen():
    DEFAULTS = GenLauncher(loglevel="error")

    # --------------
    # use argument_default=argparse.SUPPRESS so no defaults attributes are instantiated in the final Namespace
    parser = argparse.ArgumentParser(
        description="Generate upscattering events", formatter_class=argparse.ArgumentDefaultsHelpFormatter, argument_default=argparse.SUPPRESS
    )

    # BSM parameters common to all upscattering
    pt.add_common_bsm_arguments(parser, DEFAULTS)
    # Three portal model arguments
    pt.add_three_portal_arguments(parser, DEFAULTS)
    # Generic model arguments
    pt.add_generic_bsm_arguments(parser, DEFAULTS)
    # scope of the physical process of interest
    pt.add_scope_arguments(parser, DEFAULTS)
    # Monte Carlo arguments
    pt.add_mc_arguments(parser, DEFAULTS)

    kwargs = vars(parser.parse_args())

    gen_object = GenLauncher(**kwargs)
    gen_object.run(
        loglevel=kwargs.get("loglevel", gen_object.loglevel),
        verbose=kwargs.get("verbose", gen_object.verbose),
        logfile=kwargs.get("logfile", gen_object.logfile),
    )


def dn_get_examples():
    
    REPO_NAME = "DarkNews-generator"
    REPO_URL = f"https://github.com/mhostert/{REPO_NAME}.git"
    EXAMPLES_FOLDER = Path("examples")
    DESTINATION_FOLDER = Path("../DarkNews-examples")

    download_repo_cmd = f"git clone --depth 1 --filter=blob:none --sparse {REPO_URL}"
    checkout_cmd = f"git sparse-checkout set {EXAMPLES_FOLDER}"

    print("Using git clone method...")
    with subprocess.Popen([download_repo_cmd], stdout=subprocess.PIPE, shell=True) as proc:
        print(download_repo_cmd)
        print(proc.stdout.read())
    print("git clone successful...")

    os.chdir(f"{Path(REPO_NAME)}")
    
    with subprocess.Popen([checkout_cmd], stdout=subprocess.PIPE, shell=True) as proc:
        print(checkout_cmd)
        print(proc.stdout.read())
    print("git sparse-checkout successful...")

    os.rename(f"./{EXAMPLES_FOLDER}", DESTINATION_FOLDER)

    os.chdir(Path("../"))
    shutil.rmtree(REPO_NAME)

    print(f"Successfully downloaded examples folder in current directory: {DESTINATION_FOLDER}")


if __name__ == "__main__":
    dn_gen()



import logging
import sys
import numpy as np
import os
from pathlib import Path
from particle import literals as lp

# Dark Neutrino and MC stuff
import DarkNews as dn
from DarkNews import logger, prettyprinter
from DarkNews.AssignmentParser import AssignmentParser

GENERATOR_ARGS = [
    # scope
    "decay_product",
    "exp",
    "nopelastic",
    "nocoh",
    "noHC",
    "noHF",
    "nu_flavors",
    "sample_geometry",
    "make_summary_plots",
    "enforce_prompt",
    #
    # generator
    "loglevel",
    "verbose",
    "logfile",
    "neval",
    "nint",
    "neval_warmup",
    "nint_warmup",
    "seed",
    #
    # output
    "pandas",
    "parquet",
    "numpy",
    "hepevt",
    "hepevt_legacy",
    "hepmc2",
    "hepmc3",
    "hep_unweight",
    "unweighted_hep_events",
    "sparse",
    "print_to_float32",
    "path",
]

COMMON_MODEL_ARGS = [
    "name",
    "m4",
    "m5",
    "m6",
    "mzprime",
    "mhprime",
    "HNLtype",
    "mu_tr_e4",
    "mu_tr_e5",
    "mu_tr_e6",
    "mu_tr_mu4",
    "mu_tr_mu5",
    "mu_tr_mu6",
    "mu_tr_tau4",
    "mu_tr_tau5",
    "mu_tr_tau6",
    "mu_tr_44",
    "mu_tr_45",
    "mu_tr_46",
    "mu_tr_55",
    "mu_tr_56",
    "s_e4",
    "s_e5",
    "s_e6",
    "s_mu4",
    "s_mu5",
    "s_mu6",
    "s_tau4",
    "s_tau5",
    "s_tau6",
    "s_44",
    "s_45",
    "s_46",
    "s_55",
    "s_56",
    "s_66",
]

THREE_PORTAL_ARGS = [
    "gD",
    "epsilon",
    "alphaD",
    "epsilon2",
    "chi",
    "alpha_epsilon2",
    "theta",
    "Ue4",
    "Ue5",
    "Ue6",
    "Umu4",
    "Umu5",
    "Umu6",
    "Utau4",
    "Utau5",
    "Utau6",
    "UD4",
    "UD5",
    "UD6",
]

GENERIC_MODEL_ARGS = [
    "c_e4",
    "c_e5",
    "c_e6",
    "c_mu4",
    "c_mu5",
    "c_mu6",
    "c_tau4",
    "c_tau5",
    "c_tau6",
    "c_44",
    "c_45",
    "c_46",
    "c_55",
    "c_56",
    "c_66",
    "d_e4",
    "d_e5",
    "d_e6",
    "d_mu4",
    "d_mu5",
    "d_mu6",
    "d_tau4",
    "d_tau5",
    "d_tau6",
    "d_44",
    "d_45",
    "d_46",
    "d_55",
    "d_56",
    "d_66",
    "ceV",
    "ceA",
    "cuV",
    "cuA",
    "cdV",
    "cdA",
    "deV",
    "deA",
    "duV",
    "duA",
    "ddV",
    "ddA",
    "deS",
    "deP",
    "duS",
    "duP",
    "ddS",
    "ddP",
    "cprotonV",
    "cneutronV",
    "cprotonA",
    "cneutronA",
    "dprotonV",
    "dneutronV",
    "dprotonA",
    "dneutronA",
    "dprotonS",
    "dneutronS",
    "dprotonP",
    "dneutronP",
]


class GenLauncher:

    banner = r"""   ______           _        _   _                     
   |  _  \         | |      | \ | |                    
   | | | |__ _ _ __| | __   |  \| | _____      _____   
   | | | / _  | ___| |/ /   | .   |/ _ \ \ /\ / / __|  
   | |/ / (_| | |  |   <    | |\  |  __/\ V  V /\__ \  
   |___/ \__,_|_|  |_|\_\   \_| \_/\___| \_/\_/ |___/  """

    # handle parameters that can assume only certain values
    _choices = {
        "HNLtype": ["dirac", "majorana"],
        "decay_product": ["e+e-", "mu+mu-", "photon"],
        "nu_flavors": ["nu_e", "nu_mu", "nu_tau", "nu_e_bar", "nu_mu_bar", "nu_tau_bar",],
    }


    def __init__(self, param_file=None, **kwargs):
        """GenLauncher launches the generator with user-defined model parameters, experimental configs, and generation settings.

            It instantiates an object to that can run the generator for the user-defined settings.
            There are different ways to accomplish this:
                - file: read the file in input, parsing it looking for the different variables' values;
                - other arguments: variables' values.
            If a variable is not provided, it sets the default.
            at the end it looks inside the kwargs (so kwargs overwrite file definitions).

        Args:
            param_file (str, optional): path to model file. Defaults to None, which uses the parameters set at run-time.
            **kwargs: list of user-defined parameters

        Raises:
            AttributeError: when unused parameters are specified (raise only a
                warning if unused parameters are specified within the model file).
            FileNotFoundError: when param_file is specified, but the path is invalid.
            ValueError: when model, exp, or generation parameters are not well defined.
        """

        # Choose what model to initialize
        captured_3portal_args = set(THREE_PORTAL_ARGS).intersection(kwargs.keys())
        captured_generic_args = set(GENERIC_MODEL_ARGS).intersection(kwargs.keys())
        if len(captured_3portal_args) >= 0 and len(captured_generic_args) == 0:
            logger.info("Initializing the three-portal model.")
            self._model_parameters = COMMON_MODEL_ARGS + THREE_PORTAL_ARGS
            self._model_class = dn.model.ThreePortalModel
        elif len(captured_3portal_args) == 0 and len(captured_generic_args) > 0:
            logger.info("Initializing a generic model.")
            self._model_parameters = COMMON_MODEL_ARGS + GENERIC_MODEL_ARGS
            self._model_class = dn.model.GenericHNLModel
        elif len(captured_3portal_args) > 0 and len(captured_generic_args) > 0:
            logger.error(
                f"Generic Model parameters {captured_generic_args} set at the same time as {captured_3portal_args}. Cannot mix generic and three-portal model arguments together."
            )
            raise ValueError("Two types of parameters were found. You cannot mix Generic Model parameters with three-portal Model parameters.")
        else:
            raise ValueError(f"Could not determine what model type to use with kwargs.keys = {kwargs.keys()}")

        self._parameters = GENERATOR_ARGS + self._model_parameters

        # Scope parameters
        self.name = None
        self.nu_flavors = ["nu_mu"]
        self.decay_product = "e+e-"
        self.exp = "miniboone_fhc"
        self.nopelastic = False
        self.include_nelastic = False
        self.nocoh = False
        self.noHC = False
        self.noHF = False
        self.sample_geometry = False
        self.make_summary_plots = False
        self.enforce_prompt = False
        
        # Generator parameters
        self.loglevel = "INFO"
        self.verbose = False
        self.logfile = None
        self.neval = int(1e4)
        self.nint = 20
        self.neval_warmup = int(1e3)
        self.nint_warmup = 10
        self.seed = None
        
        # Output parameters
        self.pandas = True
        self.parquet = False
        self.numpy = False
        self.hepevt = False
        self.hepevt_legacy = False
        self.hepmc2 = False
        self.hepmc3 = False
        self.hep_unweight = False
        self.unweighted_hep_events = 100
        self.sparse = False
        self.print_to_float32 = False
        self.path = "."

        ####################################################
        # Loading parameters

       # the argument dictionaries (will contain valid arguments extracted from **kwargs)
        self.model_args_dict = {}

        # load file if not None
        if param_file is not None:
            self._load_file(param_file)

         # load constructor parameters
        self._load_parameters(raise_errors=True, **kwargs)

        ####################################################
        # Set up loggers
        self.prettyprinter = prettyprinter
        self.configure_logger(
            logger=logger, loglevel=self.loglevel, prettyprinter=self.prettyprinter, verbose=self.verbose, logfile=self.logfile,
        )
        prettyprinter.info(self.banner)

        ####################################################
        # Choose the model to be used in this generation
        self.bsm_model = self._model_class(**self.model_args_dict)

        ####################################################
        # Choose experiment and scope of simulation
        self.experiment = dn.detector.Detector(self.exp)

        ####################################################
        # MC evaluations and iterations
        dn.MC.NEVAL_warmup = self.neval_warmup
        dn.MC.NINT_warmup = self.nint_warmup
        dn.MC.NEVAL = self.neval
        dn.MC.NINT = self.nint

        # get the initial projectiles
        self.projectiles = [getattr(lp, nu) for nu in self.nu_flavors]

        ####################################################
        # Default data path based on model and experimental definitioons

        # set the path of the experiment name (needed in the case of custom experiment path)
        exp_path_part = os.path.basename(self.exp).rsplit(".", maxsplit=1)[0]

        _boson_string = ""
        if self.bsm_model.mzprime is not None:
            _boson_string += f"_mzprime_{self.bsm_model.mzprime:.4g}"
        if self.bsm_model.mhprime is not None:
            _boson_string += f"_mhprime_{self.bsm_model.mhprime:.4g}"

        # 3+1
        if self.bsm_model.m4 is not None and self.bsm_model.m5 is None and self.bsm_model.m6 is None:
            self.upscattered_nus = [dn.pdg.neutrino4]
            self.outgoing_nus = [dn.pdg.nulight]
            self.data_path = Path(f"{self.path}/data/{exp_path_part}/3plus1/m4_{self.bsm_model.m4:.4g}_{_boson_string}_{self.bsm_model.HNLtype}/")

        # 3+2
        elif self.bsm_model.m4 is not None and self.bsm_model.m5 is not None and self.bsm_model.m6 is None:
            ## FIXING 3+2 process chain to be nualpha --> N5 --> N4
            self.upscattered_nus = [dn.pdg.neutrino5]
            self.outgoing_nus = [dn.pdg.neutrino4]
            self.data_path = Path(
                f"{self.path}/data/{exp_path_part}/3plus2/m5_{self.bsm_model.m5:.4g}_m4_{self.bsm_model.m4:.4g}{_boson_string}_{self.bsm_model.HNLtype}/"
            )

        # 3+3
        elif self.bsm_model.m4 is not None and self.bsm_model.m5 is not None and self.bsm_model.m6 is not None:
            self.upscattered_nus = [
                dn.pdg.neutrino4,
                dn.pdg.neutrino5,
                dn.pdg.neutrino6,
            ]
            self.outgoing_nus = [dn.pdg.nulight, dn.pdg.neutrino4, dn.pdg.neutrino5]
            self.data_path = Path(
                f"{self.path}/data/{exp_path_part}/3plus3/m6_{self.bsm_model.m6:.4g}_m5_{self.bsm_model.m5:.4g}_m4_{self.bsm_model.m4:.4g}_{_boson_string}_{self.bsm_model.HNLtype}/"
            )

        else:
            logger.error("Error! Mass spectrum not allowed (m4,m5,m6) = ({self.bsm_model.m4:.4g},{self.bsm_model.m5:.4g},{self.bsm_model.m6:.4g}) GeV.")
            raise ValueError("Could not find a heavy neutrino spectrum from user input.")

        ####################################################
        # Miscellaneous checks 
        if self.hep_unweight:
            logger.warning(
                f"Unweighted events requested. This feature requires a large number of weighted events with respect to the requested number of hep-formatted events. Currently: n_unweighted/n_eval = {self.unweighted_hep_events/self.neval*100}%."
            )
        

        ####################################################
        # Create all MC cases
        self._create_all_MC_cases()

    def _load_file(self, file):
        parser = AssignmentParser({})
        try:
            parser.parse_file(file=file, comments="#")
        except FileNotFoundError:
            logger.error(f"Error! File '{file}' not found.")
            raise
        # store variables
        self._load_parameters(raise_errors=False, **parser.parameters)

    def _load_parameters(self, raise_errors=True, **kwargs):
        # start from the list of parameters available
        for parameter in self._parameters:
            # take the value from the kwargs, if not provided, go to next parameter
            try:
                value = kwargs.pop(parameter)
            except KeyError:
                continue
            
            # check the value within the choices
            # if parameter in self._choices.keys() and value not in self._choices[parameter]:
            if parameter in self._choices.keys() and [*parameter, *self._choices[parameter],] == set([*parameter, *self._choices[parameter]]):
                raise ValueError(
                    f"Parameter '{parameter}', invalid choice: {value}, (choose among " + ", ".join([f"{el}" for el in self._choices[parameter]]) + ")"
                )

            # set the parameter
            setattr(self, parameter, value)

            # save the parameters that pertain to the model classes in a dict
            if parameter in self._model_parameters:
                self.model_args_dict[parameter] = value

        # at the end, if kwargs is not empty, that would mean some parameters were unused, i.e. they are spelled wrong or do not exist: raise AttributeError
        if len(kwargs) != 0:
            if raise_errors:
                raise AttributeError("Parameters " + ", ".join(kwargs.keys()) + " were unused. Either not supported or misspelled.")
            else:
                logger.warning("Parameters " + ", ".join(kwargs.keys()) + " will not be used.")


    def _create_all_MC_cases(self, **kwargs):
        """Create MC_events objects and run the MC computations

        Args:
            **kwargs:
                FLAVORS (list):             input flavors for projectiles ['nu_e','nu_mu','nu_tau','nu_e_bar','nu_mu_bar','nu_tau_bar']
                UPSCATTERED_NUS (list):     dark NU in upscattering process
                OUTFOING_NUS (list):        output flavors
                SCATTERING_REGIMES (list):  regimes of scattering process (coherent, p-el, n-el)
                INCLUDE_HC (bool):          flag to include helicity conserving case
                INCLUDE_HF (bool):          flag to include helicity flipping case
                NO_COH (bool):              flag to skip coherent case
                NO_PELASTIC (bool):         flag to skip proton elastic case
                INCLUDE_NELASTIC (bool):         flag to skip neutron elastic case
                DECAY_PRODUCTS (list):      decay processes to include
        """

        # Default values
        scope = {
            "NO_COH": self.nocoh,
            "NO_PELASTIC": self.nopelastic,
            "INCLUDE_NELASTIC": self.include_nelastic,
            "INCLUDE_HC": not self.noHC,
            "INCLUDE_HF": not self.noHF,
            "FLAVORS": self.projectiles,
            "UPSCATTERED_NUS": self.upscattered_nus,
            "OUTGOING_NUS": self.outgoing_nus,
            "DECAY_PRODUCTS": [self.decay_product],
            "SCATTERING_REGIMES": ["coherent", "p-el"],  # , 'n-el'],
        }
        # override default with kwargs
        scope.update(kwargs)

        if not scope["SCATTERING_REGIMES"]:
            logger.error("No scattering regime found -- please specify at least one.")
            raise ValueError
        if not scope["DECAY_PRODUCTS"]:
            logger.error("No visible decay products found -- please specify at least one final state (e+e-, mu+mu- or photon).")
            raise ValueError
        if not scope["INCLUDE_HC"] and not scope["INCLUDE_HF"]:
            logger.error("No helicity structure was allowed -- please allow at least one type: HC or HF.")
            raise ValueError
        if not scope["FLAVORS"]:
            logger.error("No projectile neutrino flavors specified -- please specify at least one.")
            raise ValueError

        # create instances of all MC cases of interest
        logger.debug("Creating instances of MC cases:")
        self.gen_cases = []
        # neutrino flavor initiating scattering
        for flavor in scope["FLAVORS"]:
            # neutrino produced in upscattering
            for nu_upscattered in scope["UPSCATTERED_NUS"]:
                # neutrino produced in the subsequent decay
                for nu_outgoing in scope["OUTGOING_NUS"]:
                    # final state to consider in decay process
                    for decay_product in scope["DECAY_PRODUCTS"]:
                        # skip cases with obviously forbidden decay
                        if np.abs(nu_outgoing.pdgid) >= np.abs(nu_upscattered.pdgid):
                            continue
                        # material on which upscattering happened
                        for nuclear_target in self.experiment.NUCLEAR_TARGETS:
                            # scattering regime to use
                            for scattering_regime in scope["SCATTERING_REGIMES"]:
                                # skip disallowed regimes
                                if ((scattering_regime in ["n-el"]) and (nuclear_target.N < 1)) or (  # no neutrons
                                    (scattering_regime in ["coherent"]) and (not nuclear_target.is_nucleus)
                                ):  # coherent = p-el for hydrogen
                                    continue
                                elif (
                                    (scattering_regime in ["coherent"] and scope["NO_COH"])\
                                    or (scattering_regime in ["p-el"] and scope["NO_PELASTIC"])\
                                    or (scattering_regime in ["n-el"])\
                                    and (not scope["INCLUDE_NELASTIC"])  # do not include n-el
                                ):
                                    continue
                                else:
                                    # bundle arguments of MC_events here
                                    args = {
                                        "nuclear_target": nuclear_target,
                                        "scattering_regime": scattering_regime,
                                        "nu_projectile": flavor,
                                        "nu_upscattered": nu_upscattered,
                                        "nu_outgoing": nu_outgoing,
                                        "decay_product": decay_product,
                                        "enforce_prompt": self.enforce_prompt,
                                    }

                                    if scope["INCLUDE_HC"]:  # helicity conserving scattering
                                        self.gen_cases.append(dn.MC.MC_events(self.experiment, bsm_model=self.bsm_model, **args, helicity="conserving",))

                                    if scope["INCLUDE_HF"]:  # helicity flipping scattering
                                        self.gen_cases.append(dn.MC.MC_events(self.experiment, bsm_model=self.bsm_model, **args, helicity="flipping",))

                                    logger.debug(f"Created an MC instance of {self.gen_cases[-1].underl_process_name}.")

        return self.gen_cases

    def _scramble_df(self):
        self.df = self.df.sample(frac=1, axis=0).reset_index(drop=True)

    def _drop_zero_weight_samples(self):
        zero_entries = self.df["w_event_rate"] == 0
        if zero_entries.sum() / len(self.df.index) > 0.01:
            logger.warning(
                f"Warning: number of entries with w_event_rate = 0 surpasses 1% of number of samples. Found: {zero_entries.sum()/len(self.df.index)}."
            )
        self.df = self.df.drop(self.df[zero_entries].index).reset_index(drop=True)

    def run(self, loglevel=None, verbose=None, logfile=None, overwrite_path=None):
        """
        Run GenLauncher generation of events

        Args:
            loglevel (int, optional): what logging level to use. Can be logging.(DEBUG, INFO, WARNING, or ERROR). Defaults to logging.INFO.

            verbose (bool, optional): If true, keep date and time in the logger format. Defaults to False.

            logfile (str, optional): path to file where to log the output. Defaults to None.

            overwrite_path (str, optional): new path to save the data, it overwrites the default.

        Returns:
            pd.DataFrame: the final pandas dataframe with all the events
        """

        ####################################################
        args = {"loglevel": loglevel, "verbose": verbose, "logfile": logfile}
        for attr in args.keys():
            if args[attr] is not None:
                setattr(self, attr, args[attr])

        ############
        # temporarily overwrite path
        if overwrite_path:
            old_path = self.data_path
            self.data_path = Path(overwrite_path + "/")

        ############
        # create directory tree
        try:
            os.makedirs(self.data_path)
        except OSError:
            logger.warning("Directory tree for this run already exists. Overriding it.")

        ############
        # superseed original logger configuration
        self.configure_logger(
            logger=logger, prettyprinter=self.prettyprinter, loglevel=self.loglevel, verbose=self.verbose, logfile=self.logfile,
        )

        ######################################
        # run generator
        logger.debug("Now running the generator for each instance.")
        # Set theory params and run generation of events

        prettyprinter.info("Generating Events using the neutrino-nucleus upscattering engine")
        # numpy set used by vegas
        if self.seed:
            np.random.seed(self.seed)

        self.df = self.gen_cases[0].get_MC_events()
        for mc in self.gen_cases[1:]:
            self.df = dn.MC.get_merged_MC_output(self.df, mc.get_MC_events())

        # scramble events for minimum bias
        self._drop_zero_weight_samples()
        # eliminate events with zero weights. If too many, raise warning.
        self._scramble_df()

        #################################################
        # Save attrs
        self.df.attrs["data_path"] = self.data_path

        prettyprinter.info(
            f"Generation successful\n\nTotal events predicted:\n({np.sum(self.df['w_event_rate']):.3g} +/- {np.sqrt(np.sum(self.df['w_event_rate']**2)):.3g}) events."
        )

        ############################################################################
        # Print events to file
        self.dn_printer = dn.printer.Printer(self.df, sparse=self.sparse, print_to_float32=self.print_to_float32, decay_product=self.decay_product,)
        if self.pandas:
            self.dn_printer.print_events_to_pandas()
        if self.parquet:
            self.dn_printer.print_events_to_parquet()
        if self.numpy:
            self.dn_printer.print_events_to_ndarray()
        if self.hepevt_legacy:
            self.dn_printer.print_events_to_hepevt_legacy(hep_unweight=self.hep_unweight, unweighted_hep_events=self.unweighted_hep_events)
        if self.hepmc2:
            self.dn_printer.print_events_to_hepmc2(hep_unweight=self.hep_unweight, unweighted_hep_events=self.unweighted_hep_events)
        if self.hepmc3:
            self.dn_printer.print_events_to_hepmc3(hep_unweight=self.hep_unweight, unweighted_hep_events=self.unweighted_hep_events)
        if self.hepevt:
            self.dn_printer.print_events_to_hepevt(hep_unweight=self.hep_unweight, unweighted_hep_events=self.unweighted_hep_events)

        #############################################################################
        # Make summary plots?
        if self.make_summary_plots:
            logger.info("Making summary plots of the kinematics of the process...")
            try:
                import matplotlib
            except ImportError as e:
                logger.warning("Warning! Could not find matplotlib -- stopping the making of summary plots.")
            else:
                self.path_to_summary_plots = Path(self.data_path) / "summary_plots/"
                dn.plot_tools.batch_plot(self.df, self.path_to_summary_plots, title=rf"{self.name}")
            logger.info(f"Plots saved in {self.path_to_summary_plots}.")

        # restore overwritten path
        if overwrite_path:
            self.data_path = old_path

        return self.df

    def configure_logger(
        self, logger, loglevel=logging.INFO, prettyprinter=None, logfile=None, verbose=False,
    ):
        """
        Configure the DarkNews logger

            logger -->

            prettyprint --> secondary logger for pretty-print messages.

        Args:
            logger (logging.Logger): main DarkNews logger to be configured.
                                    It handles all debug, info, warning, and error messages

            loglevel (int, optional): what logging level to use.
                                    Can be logging.(DEBUG, INFO, WARNING, or ERROR). Defaults to logging.INFO.

            prettyprinter (logging.Logger, optional): if passed, configures this logger for the prettyprint.
                                                    Cannot override the main logger levelCannot override the main logger level.
                                                    Defaults to None.

            logfile (str, optional): path to file where to log the output. Defaults to None.

            verbose (bool, optional): If true, keep date and time in the logger format. Defaults to False.

        Raises:
            ValueError: _description_
        """

        loglevel = loglevel.upper()
        _numeric_level = getattr(logging, loglevel, None)
        if not isinstance(_numeric_level, int):
            raise ValueError("Invalid log level: %s" % self.loglevel)
        logger.setLevel(_numeric_level)

        if logfile:
            # log to files with max 1 MB with up to 4 files of backup
            handler = logging.handlers.RotatingFileHandler(f"{logfile}", maxBytes=1000000, backupCount=4)

        else:
            # stdout only
            handler = logging.StreamHandler(stream=sys.stdout)
            if prettyprinter:
                pretty_handler = logging.StreamHandler(stream=sys.stdout)
                pretty_handler.setLevel(loglevel)
                delimiter = "---------------------------------------------------------"
                pretty_handler.setFormatter(logging.Formatter(delimiter + "\n%(message)s\n"))
                # update pretty printer
                if prettyprinter.hasHandlers():
                    prettyprinter.handlers.clear()
                prettyprinter.addHandler(pretty_handler)

        handler.setLevel(loglevel)
        if verbose:
            handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(name)s:\n\t%(message)s\n", datefmt="%H:%M:%S",))
        else:
            handler.setFormatter(logging.Formatter("%(message)s"))

        if logger.hasHandlers():
            logger.handlers.clear()
        logger.addHandler(handler)




import numpy as np
from DarkNews import logger

#####################################
# particle names and props
from particle import Particle
from particle import literals as lp

################################################
# auxiliary functions -- scikit-hep particles
def in_same_doublet(p1, p2):
    if p1.pdgid in [11, 13, 15]:
        return p2.pdgid - p1.pdgid == 1 or p2.pdgid - p1.pdgid == 0
    elif p1.pdgid in [12, 14, 16]:
        return p1.pdgid - p2.pdgid == 1 or p1.pdgid - p2.pdgid == 0
    else:
        return None


def in_e_doublet(p):
    return np.abs(p.pdgid) in [11, 12]


def in_mu_doublet(p):
    return np.abs(p.pdgid) in [13, 14]


def in_tau_doublet(p):
    return np.abs(p.pdgid) in [15, 16]


def get_doublet(p):
    if in_e_doublet(p):
        return 0
    elif in_mu_doublet(p):
        return 1
    elif in_tau_doublet(p):
        return 2
    else:
        logger.error(f"Could not find doublet of {p.name}.")
        return 0


def same_doublet(p1, p2):
    return get_doublet(p1) == get_doublet(p2)


def same_particle(p1, p2):
    return p1.pdgid == p2.pdgid


def is_particle(p):
    return p.pdgid > 0


def is_antiparticle(p):
    return p.pdgid < 0


#####
# generational indices
# 0 - e, 1 -mu, 2 - tau, 3 - N4, 4 - N5, ...
def get_HNL_index(particle):
    return int(particle.name.strip("nuN")) - 1


def get_lepton_index(particle):
    return get_doublet(particle)


# mediators
photon = Particle.from_pdgid(22)

# Leptons
electron = Particle.from_pdgid(11)
positron = Particle.from_pdgid(-11)

muon = Particle.from_pdgid(13)
antimuon = Particle.from_pdgid(-13)

tau = Particle.from_pdgid(15)
antitau = Particle.from_pdgid(-15)

nue = Particle.from_pdgid(12)
nuebar = Particle.from_pdgid(-12)

numu = Particle.from_pdgid(14)
numubar = Particle.from_pdgid(-14)

nutau = Particle.from_pdgid(16)
nutaubar = Particle.from_pdgid(-16)

# Baryons
proton = Particle.from_pdgid(2212)
neutron = Particle.from_pdgid(2112)

# Nuclei
Argon40 = Particle.from_pdgid(1000180400)
Carbon12 = Particle.from_pdgid(1000060120)


########################################################################################
# Define new particles
"""
PDG code convention for new particles

As advised by the PDG, we start the new particle system with 59. The full identifier is:
    
     PDGID  =  59(particle spin code: 0-scalar 1-fermion 2-vector)(generation number)

"""


def new_particle(name, pdgid, charge=0, mass=0, **kwargs):
    return Particle(pdg_name=name, pdgid=pdgid, three_charge=charge, mass=mass, **kwargs)
    """ 
        Particle class definition:
            https://github.com/scikit-hep/particle/blob/dd3c71e0b4319f729533ff0fc2e1e8cfa49684dd/src/particle/particle/particle.py#L91
    """


"""
PDG code convention for new particles

As advised by the PDG, we start the new system with 59. The full identifier is:
    
     PDGID  =  59(particle spin code: 0-scalar 1-fermion 2-vector)(generation number)

"""

# pseudoparticle denoting all light neutrinos states
nulight = new_particle(name="nu_light", pdgid=5910, latex_name="\nu_{\rm light}")

# light neutrino mass states
neutrino1 = new_particle(name="nu1", pdgid=5911, latex_name="\nu_1")
neutrino2 = new_particle(name="nu2", pdgid=5912, latex_name="\nu_2")
neutrino3 = new_particle(name="nu3", pdgid=5913, latex_name="\nu_3")

# heavy neutrinos
neutrino4 = new_particle(name="N4", pdgid=5914, latex_name="N_4")
neutrino5 = new_particle(name="N5", pdgid=5915, latex_name="N_5")
neutrino6 = new_particle(name="N6", pdgid=5916, latex_name="N_6")

# dark photon
zprime = new_particle(name="zprime", pdgid=5921, latex_name="Z^\prime")

# three kind of new scalar particles
hprime = new_particle(name="hprime", pdgid=5901, latex_name="h^\prime")
phi = new_particle(name="phi", pdgid=5902, latex_name="\varphi")
alp = new_particle(name="alp", pdgid=5903, latex_name="a")

########################################################################################



import numpy as np
import pandas as pd
import vegas as vg

from DarkNews import logger, prettyprinter

from collections import defaultdict
from functools import partial

from DarkNews import processes
from DarkNews import integrands
from DarkNews import const
from DarkNews import pdg
from DarkNews import geom

NINT = 10
NEVAL = 1000
NINT_warmup = 10
NEVAL_warmup = 1000


class MC_events:
    """MC events generated with importance sampling (vegas)

    Correct weights are computed from cross-section, decay width,
    and experimental considerations

    Args:
        experiment:             instance of Detector class
        target:                 scattering target class
        scattering_regime:      regime for scattering process, choices: ['coherent', 'p-el', 'n-el', 'DIS'] (future: RES)
        nu_projectile:          incoming neutrino flavor
        nu_upscattered:         intermediate dark NU in upscattering process
        nu_outgoing:            outgoing neutrino flavor
        decay_product:          visible decay products in the detector
        helicity:               helicity of the up-scattered neutrino
    """

    def __init__(self, experiment, bsm_model, **kwargs):

        # default parameters
        scope = {
            "nu_projectile": pdg.numu,
            "nu_upscattered": pdg.neutrino4,
            "nu_outgoing": pdg.nulight,
            "scattering_regime": "coherent",
            "decay_product": ["e+e-"],
            "helicity": "conserving",
            "enforce_prompt": False,
        }

        scope.update(kwargs)
        self.scope = scope
        self.experiment = experiment

        # set target properties for this scattering regime
        if "nuclear_target" in scope:
            self.nuclear_target = scope["nuclear_target"]
        else:
            self.nuclear_target = self.experiment.NUCLEAR_TARGETS[0]
            logger.warning("No target passed to MC_events, using first entry in experiment class instead.")

        # identifiers for particles in the process
        self.nu_projectile = scope["nu_projectile"]
        self.nu_upscattered = scope["nu_upscattered"]
        self.nu_outgoing = scope["nu_outgoing"]
        self.helicity = scope["helicity"]

        if scope["decay_product"] == "e+e-":
            self.decay_product = pdg.electron
            # process being considered
            DECAY_PRODUCTS = f"{self.decay_product.invert().name} {self.decay_product.name}"
            self.decays_to_dilepton = True
            self.decays_to_singlephoton = False

        elif scope["decay_product"] == "mu+mu-":
            self.decay_product = pdg.muon
            # process being considered
            DECAY_PRODUCTS = f"{self.decay_product.invert().name} {self.decay_product.name}"
            self.decays_to_dilepton = True
            self.decays_to_singlephoton = False

        elif scope["decay_product"] == "photon":
            self.decay_product = pdg.photon
            DECAY_PRODUCTS = f"{self.decay_product.name}"
            self.decays_to_dilepton = False
            self.decays_to_singlephoton = True

        else:
            logger.error(f"Error! Could not find decay product: {scope['decay_product']}")
            raise ValueError

        self.bsm_model = bsm_model
        # scope for upscattering process
        self.ups_case = processes.UpscatteringProcess(
            nu_projectile=self.nu_projectile,
            nu_upscattered=self.nu_upscattered,
            nuclear_target=self.nuclear_target,
            scattering_regime=self.scope["scattering_regime"],
            helicity=self.helicity,
            TheoryModel=bsm_model,
        )

        if self.decays_to_dilepton:
            # scope for decay process
            self.decay_case = processes.FermionDileptonDecay(
                nu_parent=self.nu_upscattered,
                nu_daughter=self.nu_outgoing,
                final_lepton1=self.decay_product,
                final_lepton2=self.decay_product,
                h_parent=self.ups_case.h_upscattered,
                TheoryModel=bsm_model,
            )
        elif self.decays_to_singlephoton:
            # scope for decay process
            self.decay_case = processes.FermionSinglePhotonDecay(
                nu_parent=self.nu_upscattered, nu_daughter=self.nu_outgoing, h_parent=self.ups_case.h_upscattered, TheoryModel=bsm_model,
            )
        else:
            logger.error("Error! Could not determine what type of decay class to use.")
            raise ValueError

        if self.ups_case.Ethreshold > self.experiment.ERANGE[-1]:
            logger.error(f"Particle {self.nu_upscattered.name} is too heavy to be produced in the energy range of experiment {self.experiment.NAME}")
            raise ValueError

        if self.ups_case.m_ups != self.decay_case.m_parent:
            logger.error(
                f"Error! Mass of HNL produced in neutrino scattering m_ups = {self.ups_case.m_upscattered} GeV, \
                    different from that of parent HNL, m_parent = {self.decay_case.m_parent} GeV."
            )
            raise ValueError

        # process being considered
        self.underl_process_name = \
            f"{self.nu_projectile.name} {self.ups_case.target.name} --> \
{self.nu_upscattered.name}  {self.ups_case.target.name} --> \
{self.nu_outgoing.name} {DECAY_PRODUCTS} {self.ups_case.target.name}"

    def get_total_xsec(self, Enu):
        """
        Returns the total upscattering xsec for a fixed neutrino energy

        """

        DIM = 1
        self.Enu = Enu

        # below threshold
        if Enu < (self.ups_case.Ethreshold):
            return 0.0

        batch_f = integrands.UpscatteringXsec(dim=DIM, Enu=self.Enu, MC_case=self)
        integ = vg.Integrator(DIM * [[0.0, 1.0]])  # unit hypercube

        integrals = run_vegas(batch_f, integ, NINT=NINT, NEVAL=NEVAL, NINT_warmup=NINT_warmup, NEVAL_warmup=NEVAL_warmup,)
        logger.debug("Main VEGAS run completed.")

        #############
        # integrated xsec coverted to cm^2
        tot_xsec = integrals["diff_xsec"].mean * const.attobarn_to_cm2 * batch_f.norm["diff_xsec"]

        # How many constituent targets inside scattering regime?
        if self.scope["scattering_regime"] == "coherent":
            target_multiplicity = 1
        elif self.scope["scattering_regime"] == "p-el":
            target_multiplicity = self.nuclear_target.Z
        elif self.scope["scattering_regime"] == "n-el":
            target_multiplicity = self.nuclear_target.N
        elif self.scope["scattering_regime"] == "DIS":
            target_multiplicity = 1
        else:
            logger.error(f"Scattering regime {self.scope['scattering_regime']} not supported.")

        logger.debug("Total cross section calculated.")
        return tot_xsec * target_multiplicity

    def get_MC_events(self):
        """
        Returns MC events from importance sampling

        Handles the off and on shell cases separately

        After sampling the integrands, we normalize the integral by
        the integral of the decay rates, which is also estimated by VEGAS
        in the multi-component integrand.

        The first integrand entry is the one VEGAS uses to optimize the importance sampling.
        """

        logger.info(f"{self.underl_process_name}")
        logger.info(f"Helicity {self.helicity} upscattering.")

        # ###############################
        # Some experimental definitions
        self.flux = self.experiment.neutrino_flux(self.nu_projectile)
        self.EMIN = max(self.experiment.ERANGE[0], 1.05 * self.ups_case.Ethreshold)
        self.EMAX = self.experiment.ERANGE[1]

        if self.decays_to_dilepton:

            if self.decay_case.on_shell:
                DIM = 3
                logger.info(f"{self.nu_upscattered.name} decays via on-shell Z'.")
            elif self.decay_case.off_shell:
                DIM = 6
                logger.info(f"{self.nu_upscattered.name} three-body decays.")
        elif self.decays_to_singlephoton:
            DIM = 3
            logger.info(f"{self.nu_upscattered.name} decays via TMM.")
        else:
            logger.error("ERROR! Could not determine decay process.")
            raise ValueError("Could not determine decay process.")
        integrand_type = integrands.UpscatteringHNLDecay

        #########################################################################
        # BATCH SAMPLE INTEGRAND OF INTEREST
        logger.debug(f"Running VEGAS for DIM={DIM}")
        batch_f = integrand_type(dim=DIM, Emin=self.EMIN, Emax=self.EMAX, MC_case=self)
        integ = vg.Integrator(DIM * [[0.0, 1.0]])  # unit hypercube
        result = run_vegas(batch_f, integ, NINT=NINT, NEVAL=NEVAL, NINT_warmup=NINT_warmup, NEVAL_warmup=NEVAL_warmup,)
        logger.debug("Main VEGAS run completed.")

        logger.debug(f"Vegas results for the integrals: {result.summary()}")
        ##########################################################################

        #########################################################################
        # GET THE INTEGRATION SAMPLES and translate to physical variables in MC events
        samples, weights = get_samples(integ, batch_f)
        tot_nevents = len(weights["diff_event_rate"])

        logger.debug(f"Normalization factors in MC: {batch_f.norm}.")
        logger.debug(f"Vegas results for diff_event_rate: {weights['diff_event_rate'].sum()}")
        logger.debug(f"Vegas results for diff_flux_avg_xsec: {weights['diff_flux_avg_xsec'].sum()}")

        four_momenta = integrands.get_momenta_from_vegas_samples(samples, MC_case=self)
        ##########################################################################

        ##########################################################################
        # SAVE ALL EVENTS AS A PANDAS DATAFRAME
        particles = list(four_momenta.keys())
        columns = particles + ["pos_scatt"]
        indexing = [columns, ["0", "1", "2", "3"]]
        columns_index = pd.MultiIndex.from_product(indexing)

        # create auxiliary data list
        aux_data = []
        # create pandas dataframe
        for p in particles:
            for component in range(4):
                aux_data.append(four_momenta[p][:, component])
        # scattering position
        aux_data.append(np.zeros((tot_nevents,)))
        aux_data.append(np.zeros((tot_nevents,)))
        aux_data.append(np.zeros((tot_nevents,)))
        aux_data.append(np.zeros((tot_nevents,)))

        df_gen = pd.DataFrame(np.stack(aux_data, axis=-1), columns=columns_index)

        # differential weights
        for column in df_gen:
            if "w_" in column:
                df_gen[column, ""] = df_gen[column]

        # Normalize weights and total integral with decay rates and set units to nus*cm^2/POT
        decay_rates = 1
        for decay_step in (k for k in batch_f.int_dic.keys() if "decay_rate" in k):
            logger.debug(f"Vegas results for {decay_step}: {weights[decay_step].sum()}")

            # saving decay weights and integrals
            df_gen[f"w_{decay_step}".replace("diff_", "")] = weights[decay_step] * batch_f.norm[decay_step]

            # combining all decay rates into one factor
            decay_rates *= df_gen[f"w_{decay_step}".replace("diff_", "")].sum()

        # How many constituent targets inside scattering regime?
        if self.scope["scattering_regime"] == "coherent":
            target_multiplicity = 1
        elif self.scope["scattering_regime"] == "p-el":
            target_multiplicity = self.nuclear_target.Z
        elif self.scope["scattering_regime"] == "n-el":
            target_multiplicity = self.nuclear_target.N
        elif self.scope["scattering_regime"] == "DIS":
            target_multiplicity = 1
        else:
            logger.error(f"Scattering regime {self.scope['scattering_regime']} not supported.")

        # Normalize to total exposure
        exposure = self.experiment.NUMBER_OF_TARGETS[self.nuclear_target.name] * self.experiment.POTS

        # differential rate weights
        df_gen["w_event_rate"] = (
            weights["diff_event_rate"] * const.attobarn_to_cm2 / decay_rates * target_multiplicity * exposure * batch_f.norm["diff_event_rate"]
        )

        # flux averaged xsec weights (neglecting kinematics of decay)
        df_gen["w_flux_avg_xsec"] = weights["diff_flux_avg_xsec"] * const.attobarn_to_cm2 * target_multiplicity * exposure * batch_f.norm["diff_flux_avg_xsec"]

        df_gen.insert(
            loc=len(df_gen.columns), column="target", value=np.full(tot_nevents, self.ups_case.target.name),
        )
        df_gen["target"] = df_gen["target"].astype("string")
        df_gen.insert(
            loc=len(df_gen.columns), column="target_pdgid", value=self.ups_case.target.pdgid,
        )
        df_gen["target_pdgid"] = df_gen["target_pdgid"].astype("int")
        df_gen.insert(
            loc=len(df_gen.columns), column="scattering_regime", value=np.full(tot_nevents, self.ups_case.scattering_regime),
        )
        df_gen["scattering_regime"] = df_gen["scattering_regime"].astype("string")
        df_gen.insert(
            loc=len(df_gen.columns), column="helicity", value=np.full(tot_nevents, self.helicity),
        )
        df_gen["helicity"] = df_gen["helicity"].astype("string")
        df_gen.insert(
            loc=len(df_gen.columns), column="underlying_process", value=np.full(tot_nevents, self.underl_process_name),
        )
        df_gen["underlying_process"] = df_gen["underlying_process"].astype("string")

        # Helicity of incoming neutrino
        if self.nu_projectile.pdgid < 0:
            df_gen.insert(
                loc=len(df_gen.columns), column="h_projectile", value=np.full(tot_nevents, +1),
            )
        elif self.nu_projectile.pdgid > 0:
            df_gen.insert(
                loc=len(df_gen.columns), column="h_projectile", value=np.full(tot_nevents, -1),
            )

        # Helicity of outgoing HNL
        if self.helicity == "conserving":
            df_gen["h_parent", ""] = df_gen["h_projectile"]
        elif self.helicity == "flipping":
            df_gen["h_parent", ""] = -df_gen["h_projectile"]

        # saving the experiment class
        df_gen.attrs["experiment"] = self.experiment

        # saving the bsm_model class
        df_gen.attrs["model"] = self.bsm_model

        # saving the lifetime of the parent (upscattered) HNL
        df_gen.attrs[f"{self.nu_upscattered.name}_ctau0"] = const.get_decay_rate_in_cm(df_gen["w_decay_rate_0"].sum())

        # #########################################################################
        # PROPAGATE PARENT PARTICLE

        self.experiment.set_geometry()
        self.experiment.place_scatters(df_gen)

        if self.scope["enforce_prompt"]:
            geom.place_decay(df_gen, "P_decay_N_parent", l_decay_proper_cm=0.0, label="pos_decay")
        else:
            # decay only the first mother (typically the HNL produced)
            logger.info(f"Parent {self.ups_case.nu_upscattered.name} proper decay length: {df_gen.attrs[f'{self.nu_upscattered.name}_ctau0']:.3E} cm.\n")
            geom.place_decay(
                df_gen, "P_decay_N_parent", l_decay_proper_cm=df_gen.attrs[f"{self.nu_upscattered.name}_ctau0"], label="pos_decay",
            )

        # print final result
        logger.info(f"Predicted ({df_gen['w_event_rate'].sum():.3g} +/- {np.sqrt((df_gen['w_event_rate']**2).sum()):.3g}) events.\n")

        return df_gen


# merge all generation cases into one dictionary
def get_merged_MC_output(df1, df2):
    """
    take two pandas dataframes with events and combine them.
    Resetting index to go from (0,n_1+n_2) where n_i is the number of events in dfi
    """
    logger.debug(f"Appending {df2.underlying_process[0]}")
    df = pd.concat([df1, df2], axis=0).reset_index(drop=True)

    # for older versions of pandas, concat does not keep the attributes
    #  -- if they disappear, force first dataframe.
    if not df.attrs:
        logger.debug("DEBUG: Forcing the storage of the df.attrs using the first dataframe. This is done automatically for newer versions of pandas.")
        df.attrs = df1.attrs

    # Now we merge lifetimes
    for i in range(4, 7):
        this_ctau0 = f"N{i}_ctau0"
        if this_ctau0 in df1.attrs.keys() and this_ctau0 in df2.attrs.keys():
            # take the average
            df.attrs[this_ctau0] = 0.5 * (df1.attrs[this_ctau0] + df2.attrs[this_ctau0])

    return df


def get_samples(integ, batch_integrand, return_jac=False):
    """_summary_

    Args:
        integ (vegas.Integrator): vegas integrator object initialized by the user.

        batch_integrand (vegas.BatchIntegrand): vegas batch_integrand object created by the user.
        These are defined in integrands.py

        return_jac (bool, optional): if True, returns the jacobian of the integrand as well. Defaults to False.

    Raises:
        ValueError: if the integrand evaluates to nan

    Returns:
        tuple of np.ndarrays:
    """

    unit_samples = batch_integrand.dim * [[]]
    weights = defaultdict(partial(np.ndarray, 0))

    for x, y, wgt in integ.random_batch(yield_y=True, fcn=batch_integrand):

        # compute integrand on samples including jacobian factors
        if integ.uses_jac:
            jac = integ.map.jac1d(y)
        else:
            jac = None

        fx = batch_integrand(x, jac=jac)
        # weights
        for fx_i in fx.keys():
            if np.any(np.isnan(fx[fx_i])):
                raise ValueError(f"integrand {fx_i} evaluates to nan")
            weights[fx_i] = np.append(weights[fx_i], wgt * fx[fx_i])

        # MC samples in unit hypercube
        for i in range(batch_integrand.dim):
            unit_samples[i] = np.append(unit_samples[i], x[:, i])

    if return_jac:
        return np.array(unit_samples), weights, jac
    else:
        return np.array(unit_samples), weights


def run_vegas(batch_f, integ, NINT=10, NEVAL=1000, NINT_warmup=10, NEVAL_warmup=1000, **kwargs):
    """
    Function that calls vegas evaluations. This function defines the vegas parameters used by
    DarkNews throughout.

    Args:
        integ (vegas.Integrator): vegas integrator object initialized by the user.

        batch_integrand (vegas.BatchIntegrand): vegas batch_integrand object created by the user.

        NINT (int, optional): number of iterations for vegas. Defaults to 10.
        NEVAL (int, optional): number of evaluations in an iteration for vegas to 1000.
        NINT_warmup (int, optional): same as above, but for warmup run. Defaults to 10.
        NEVAL_warmup (int, optional): same as above, but for warmup run. Defaults to 1000.

    Returns:
        integ (vegas.Integrator): with the evaluated integrals.
    """

    # warm up the MC, adapting to the integrand
    integ(batch_f, nitn=NINT_warmup, neval=NEVAL_warmup, uses_jac=True, **kwargs)
    logger.debug("VEGAS warm-up completed.")

    # sample again, now saving result and turning off further adaption
    return integ(batch_f, nitn=NINT, neval=NEVAL, uses_jac=True, **kwargs)



import os
import numpy as np
from scipy.interpolate import splprep, splev

from DarkNews import const
from DarkNews import fourvec as fv

import matplotlib

import matplotlib.pyplot as plt
from matplotlib import rc, rcParams

# from matplotlib.pyplot import *
from matplotlib.pyplot import cm

###########################
fsize = 12
fsize_annotate = 10

std_figsize = (1.2 * 3.7, 1.6 * 2.3617)
std_axes_form = [0.16, 0.15, 0.81, 0.76]

# standard figure
def std_fig(ax_form=std_axes_form, figsize=std_figsize, rasterized=False):

    rcparams = {
        "axes.labelsize": fsize,
        "xtick.labelsize": fsize,
        "ytick.labelsize": fsize,
        "figure.figsize": std_figsize,
        "legend.frameon": False,
        "legend.loc": "best",
    }
    plt.rcParams["text.latex.preamble"] = r"\usepackage{amsmath}\usepackage{amssymb}"
    rc("font", **{"family": "serif", "serif": ["Computer Modern"]})
    rc("text", usetex=True)
    rcParams.update(rcparams)
    matplotlib.rcParams["hatch.linewidth"] = 0.3
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes(ax_form, rasterized=rasterized)
    ax.patch.set_alpha(0.0)

    return fig, ax


# standard saving function
def std_savefig(fig, path, dpi=400, **kwargs):
    fig.savefig(path, dpi=dpi, **kwargs)


def sci_notation(num, decimal_digits=1, precision=None, exponent=None):
    """
    Returns a string representation of the scientific
    notation of the given number formatted for use with
    LaTeX or Mathtext, with specified number of significant
    decimal digits and precision (number of decimal digits
    to show). The exponent to be used can also be specified
    explicitly.
    """
    if num != 0:
        if exponent is None:
            exponent = int(np.floor(np.log10(abs(num))))
        coeff = round(num / float(10 ** exponent), decimal_digits)
        if precision is None:
            precision = decimal_digits

        return r"${0:.{2}f}\cdot10^{{{1:d}}}$".format(coeff, exponent, precision)
    else:
        return r"0"


def get_hist(ax):
    n, bins = [], []
    for rect in ax.patches:
        ((x0, y0), (x1, y1)) = rect.get_bbox().get_points()
        n.append(y1 - y0)
        bins.append(x0)  # left edge of each bin
    bins.append(x1)  # also get right edge of last bin
    return n, bins


def histogram1D(plotname, obs, w, XLABEL, TITLE, nbins, regime=None, colors=None, legends=None, rasterized=True, TMIN=None, TMAX=None):

    fig, ax = std_fig()
    ax.patch.set_alpha(0.0)

    if TMIN is None:
        TMIN = np.min(obs)
    if TMAX is None:
        TMAX = np.max(obs)

    nregimes = len(regime)
    bin_e = np.linspace(TMIN, TMAX, nbins + 1, endpoint=True)
    bin_w = bin_e[1:] - bin_e[:-1]
    if regime and not legends:
        legends = [f"case {i}" for i in range(nregimes)]
    if regime and not colors:
        color = cm.rainbow(np.linspace(0, 1, nregimes))
        colors = [c for c in color]

    if regime:
        htotal = np.zeros((nbins))
        nregimes = len(regime)
        for i in range(np.shape(regime)[0]):
            case = regime[i]
            h, bins = np.histogram(obs[case], weights=w[case], bins=bin_e)

            ax.bar(
                bins[:-1], h, bottom=htotal, label=legends[i], width=bin_w,
                ec=None, facecolor=colors[i], alpha=1, align="edge", lw=0.0,
                rasterized=rasterized
            )

            htotal += h
            ax.step(np.append(bins[:-1], 10e10), np.append(htotal, 0.0), where="post",
            c="black", lw=0.25, rasterized=rasterized)

    else:
        h, bins = np.histogram(obs, weights=w, bins=nbins, range=(TMIN, TMAX))
        ax.bar(bins[:-1], h, width=bin_w, ec=None, fc="indigo", alpha=0.8, align="edge", lw=0.0, rasterized=rasterized)


    ax.set_title(TITLE, fontsize=0.8 * fsize)
    ax.legend(frameon=False, loc="best")
    ax.set_xlabel(XLABEL, fontsize=fsize)
    ax.set_ylabel(r"events", fontsize=fsize)

    if TMIN != TMAX:
        ax.set_xlim(TMIN, TMAX)
    elif TMIN == TMAX and TMIN != 0.0:
        ax.set_xlim(TMIN * 0.5, TMIN * 2)

    ax.set_ylim(0.0, ax.get_ylim()[1] * 1.1)
    std_savefig(fig, plotname, dpi=400)
    plt.close()


def histogram2D(plotname, obsx, obsy, w, xrange=None, yrange=None, xlabel="x", ylabel="y", title="Dark News", nbins=20, logx=False, logy=False):
    fsize = 11
    fig, ax = std_fig(ax_form=[0.15, 0.15, 0.78, 0.74])

    if logx:
        obsx = np.log10(obsx)
    if logy:
        obsy = np.log10(obsy)

    if not xrange:
        xrange = [np.min(obsx), np.max(obsx)]
    if not yrange:
        yrange = [np.min(obsy), np.max(obsy)]

    bar = ax.hist2d(obsx, obsy, bins=nbins, weights=w, range=[xrange, yrange], cmap="Blues", density=True)

    ax.set_title(title, fontsize=fsize)
    cbar_R = fig.colorbar(bar[3], ax=ax)
    # cbar_R.ax.set_ylabel(r"a.u.", rotation=90)
    cbar_R.ax.set_ylabel(r"events", rotation=90)

    ax.set_xlabel(xlabel, fontsize=fsize)
    ax.set_ylabel(ylabel, fontsize=fsize)
    std_savefig(fig, plotname, dpi=400)
    plt.close()


def batch_plot(df, PATH, title="Dark News"):
    # regimes
    coherent = df["scattering_regime"] == "coherent"
    pel = df["scattering_regime"] == "p-el"
    HC = df["helicity"] == "conserving"
    HF = df["helicity"] == "flipping"
    cases = [coherent & HC, pel & HC, coherent & HF, pel & HF]
    # case_names = [r"coherent conserving", r"p-el conserving", r"coherent flipping", r"p-el flipping"]
    case_shorthands = [r"coh HC", r"p-el HC", r"coh HF", r"p-el HF"]
    colors = ["dodgerblue", "lightblue", "violet", "pink"]
    args = {"regime": cases, "colors": colors, "legends": case_shorthands, "rasterized": False}

    if not os.path.exists(PATH):
        os.mkdir(PATH)

    # weights
    w = df["w_event_rate", ""]
    w_pel = df["w_event_rate", ""][pel]
    w_coh = df["w_event_rate", ""][coherent]

    df["E_N"] = df["P_decay_N_parent", "0"]
    df["E_Had"] = df["P_recoil", "0"]
    df["M_had"] = fv.df_inv_mass(df["P_recoil"], df["P_recoil"])
    df["Q2"] = -(2 * df["M_had"] ** 2 - 2 * df["E_Had"] * df["M_had"])

    df["costheta_N"] = fv.df_cos_azimuthal(df["P_decay_N_parent"])
    df["costheta_nu"] = fv.df_cos_azimuthal(df["P_decay_N_daughter"])
    df["costheta_Had"] = fv.df_cos_azimuthal(df["P_recoil"])
    df["theta_N"] = np.arccos(df["costheta_N"]) * const.rad_to_deg
    df["theta_nu"] = np.arccos(df["costheta_nu"]) * const.rad_to_deg

    df["theta_proton"] = np.arccos(df["costheta_Had"][pel]) * const.rad_to_deg
    df["theta_nucleus"] = np.arccos(df["costheta_Had"][coherent]) * const.rad_to_deg

    df["T_proton"] = (df["E_Had"] - df["M_had"])[pel]
    df["T_nucleus"] = (df["E_Had"] - df["M_had"])[coherent]

    histogram1D(PATH / "1D_E_nu_truth.pdf", df["P_projectile", "0"], w, r"$E_\nu^{\rm truth}/$GeV", title, 20, **args)
    histogram1D(PATH / "1D_E_N.pdf", df["E_N"], w, r"$E_N/$GeV", title, 20, **args)
    histogram1D(PATH / "1D_cost_nu.pdf", df["costheta_nu"], w, r"$\cos(\theta_{\nu_{\rm proj} \nu_{\rm out}})$", title, 40, **args)
    histogram1D(PATH / "1D_theta_nu.pdf", df["theta_nu"], w, r"$\theta_{\nu_{\rm proj} \nu_{\rm out}}$", title, 40, **args)
    histogram1D(PATH / "1D_costN.pdf", df["costheta_N"], w, r"$\cos(\theta_{\nu_{\rm proj} N})$", title, 20, **args)

    histogram1D(PATH / "1D_Q.pdf", np.sqrt(df["Q2"]), w, r"$Q/$GeV", title, 50, **args)
    histogram1D(PATH / "1D_Q2.pdf", df["Q2"], w, r"$Q^2/$GeV$^2$", title, 50, **args)
    histogram1D(PATH / "1D_T_proton.pdf", df["T_proton"][pel] * 1e3, w_pel, r"$T_{\rm p^+}$/MeV", "el proton only", 50, **args)
    histogram1D(PATH / "1D_theta_proton.pdf", df["theta_proton"][pel], w_pel, r"$\theta_{p^+}$ degrees", "el proton only", 50, **args)
    histogram1D(PATH / "1D_T_nucleus.pdf", df["T_nucleus"][coherent] * 1e3, w_coh, r"$T_{\rm Nucleus}$/MeV", "coh nucleus only", 50, **args)
    histogram1D(PATH / "1D_theta_nucleus.pdf", df["theta_nucleus"][coherent], w_coh, r"$\theta_{\rm Nucleus}$ degrees", "coh nucleus only", 50, **args)

    # Variables of interest for single photon decays
    if "P_decay_photon" in df.columns.levels[0]:
        # some useful definitions for four momenta

        # variables
        df["E_photon"] = df["P_decay_photon", "0"]

        df["costheta_photon"] = fv.df_cos_azimuthal(df["P_decay_photon"])
        df["theta_photon"] = np.arccos(df["costheta_photon"]) * const.rad_to_deg

        df["costheta_photon_had"] = fv.df_cos_opening_angle(df["P_decay_photon"], df["P_recoil"])
        df["theta_photon_had"] = np.arccos(df["costheta_photon_had"]) * const.rad_to_deg

        # CCQE neutrino energy
        df["E_nu_reco"] = const.m_proton * df["E_photon"] / (const.m_proton - df["E_photon"] * (1.0 - df["costheta_photon"]))

        # HISTOGRAM 2D ##################################################
        n2D = 40
        args_2d = {"title": title, "nbins": n2D}

        histogram2D(
            PATH / "2D_EN_Ephoton.pdf",
            df["E_N"],
            df["E_photon"],
            w,
            xrange=[0.0, 2.0],
            yrange=[0.0, 2.0],
            xlabel=r"$E_{N}$/GeV",
            ylabel=r"$E_\gamma$/GeV",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_Ehad_Ephoton.pdf",
            df["T_proton"][pel] * 1e3,
            df["E_photon"][pel],
            w_pel,
            xrange=[0.0, 1000],
            yrange=[0.0, 2.0],
            xlabel=r"$T_{\rm proton}$/MeV",
            ylabel=r"$(E_{\ell^+} + E_{\ell^-})$/GeV",
            title=title + " proton-elastic only",
            nbins=n2D,
        )

        # HISTOGRAMS 1D ####################################################
        # momentum exchange
        histogram1D(PATH / "1D_E_photon.pdf", df["E_photon"], w, r"$E_\gamma$/GeV", title, 100, TMIN=0.0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_E_nu_QEreco.pdf", df["E_nu_reco"], w, r"$E_\nu^{\rm QE-reco}/$GeV", title, 40, TMIN=0.0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_cost_photon.pdf", df["costheta_photon"], w, r"$\cos(\theta_{\gamma})$", title, 40, TMIN=-1, TMAX=1, **args)
        histogram1D(PATH / "1D_cost_photon_had.pdf", df["costheta_photon_had"], w, r"$\cos(\theta_{\gamma {\rm hadron}})$", title, 40, TMIN=-1, TMAX=1, **args)
        histogram1D(PATH / "1D_theta_photon.pdf", df["theta_photon"], w, r"$\theta_{\gamma}$ degrees", title, 40, TMIN=0.0, TMAX=90, **args)
        histogram1D(PATH / "1D_theta_photon_had.pdf", df["theta_photon_had"], w, r"$\theta_{\gamma  {\rm hadron}}$ degrees", title, 40, TMIN=0.0, **args)

    # Variables of interest for dilepton decays
    if "P_decay_ell_minus" in df.columns.levels[0]:
        # some useful definitions for four momenta
        for i in range(4):
            df["P_decay_ellell", i] = df["P_decay_ell_minus", f"{i}"] + df["P_decay_ell_plus", f"{i}"]

        # variables
        df["E_Z"] = df["P_decay_ell_minus", "0"] + df["P_decay_ell_plus", "0"]
        df["E_lp"] = df["P_decay_ell_plus", "0"]
        df["E_lm"] = df["P_decay_ell_minus", "0"]
        df["E_tot"] = df["E_lm"] + df["E_lp"]
        df["E_asy"] = (df["E_lp"] - df["E_lm"]) / (df["E_lp"] + df["E_lm"])

        df["inv_mass"] = fv.df_inv_mass(df["P_decay_ellell"], df["P_decay_ellell"])

        df["costheta_sum"] = fv.df_cos_azimuthal(df["P_decay_ellell"])
        df["costheta_lp"] = fv.df_cos_azimuthal(df["P_decay_ell_plus"])
        df["costheta_lm"] = fv.df_cos_azimuthal(df["P_decay_ell_minus"])

        df["costheta_sum_had"] = fv.df_cos_opening_angle(df["P_decay_ellell"], df["P_recoil"])
        df["theta_sum_had"] = np.arccos(df["costheta_sum_had"]) * const.rad_to_deg

        df["theta_sum"] = np.arccos(df["costheta_sum"]) * const.rad_to_deg
        df["theta_lp"] = np.arccos(df["costheta_lp"]) * const.rad_to_deg
        df["theta_lm"] = np.arccos(df["costheta_lm"]) * const.rad_to_deg

        df["Delta_costheta"] = fv.df_cos_opening_angle(df["P_decay_ell_minus"], df["P_decay_ell_plus"])
        df["Delta_theta"] = np.arccos(df["Delta_costheta"]) * const.rad_to_deg

        minus_lead = df["P_decay_ell_minus", "0"] >= df["P_decay_ell_plus", "0"]
        plus_lead = df["P_decay_ell_minus", "0"] < df["P_decay_ell_plus", "0"]

        df["E_subleading"] = np.minimum(df["P_decay_ell_minus", "0"], df["P_decay_ell_plus", "0"])
        df["E_leading"] = np.maximum(df["P_decay_ell_minus", "0"], df["P_decay_ell_plus", "0"])

        df["theta_subleading"] = df["theta_lp"] * plus_lead + df["theta_lm"] * minus_lead
        df["theta_leading"] = df["theta_lp"] * (~plus_lead) + df["theta_lm"] * (~minus_lead)

        # CCQE neutrino energy
        _denominator = const.m_proton - df["E_tot"] * (1.0 - (df["costheta_lm"] * df["E_lm"] + df["costheta_lp"] * df["E_lp"]) / df["E_tot"])
        df["E_nu_reco"] = const.m_proton * df["E_tot"] / _denominator

        # HISTOGRAM 2D ##################################################
        n2D = 40
        args_2d = {"title": title, "nbins": n2D}

        histogram2D(
            PATH / "2D_EN_Etot.pdf",
            df["E_N"],
            df["E_tot"],
            w,
            xrange=[0.0, 2.0],
            yrange=[0.0, 2.0],
            xlabel=r"$E_{N}$/GeV",
            ylabel=r"$(E_{\ell^-}+E_{\ell^+})$/GeV",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_Ep_Em.pdf",
            df["E_lm"],
            df["E_lp"],
            w,
            xrange=[0.0, 2.0],
            yrange=[0.0, 2.0],
            xlabel=r"$E_{\ell^-}$/GeV",
            ylabel=r"$E_{\ell^+}$/GeV",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_dtheta_Etot.pdf",
            df["Delta_theta"],
            df["E_tot"],
            w,
            xrange=[0.0, 90],
            yrange=[0.0, 2.0],
            xlabel=r"$\Delta \theta_{\ell \ell}$ degrees",
            ylabel=r"$(E_{\ell^+}+E_{\ell^-})$/GeV",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_Easyabs_dtheta.pdf",
            np.abs(df["E_asy"]),
            df["Delta_theta"],
            w,
            xrange=[0.0, 1.0],
            yrange=[0.0, 90.0],
            xlabel=r"$|E_{\ell^+}-E_{\ell^-}|$/($E_{\ell^+}+E_{\ell^-}$)",
            ylabel=r"$\Delta \theta_{\ell \ell}$ degrees",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_Easyabs_theta_sum.pdf",
            np.abs(df["E_asy"]),
            df["theta_sum"],
            w,
            xrange=[0.0, 1.0],
            yrange=[0.0, 40.],
            xlabel=r"$|E_{\rm asy}|$",
            ylabel=r"$\theta_{(\ell \ell)\nu_{\rm proj}}$/GeV",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_Easyabs_Etot.pdf",
            np.abs(df["E_asy"]),
            df["E_tot"],
            w,
            xrange=[0.0, 1.0],
            yrange=[0.0, 2.0],
            xlabel=r"$|E_{\rm asy}|$",
            ylabel=r"$(E_{\ell^+}+E_{\ell^-})$/GeV",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_Ehad_Etot.pdf",
            df["T_proton"][pel] * 1e3,
            df["E_tot"][pel],
            w_pel,
            xrange=[0.0, 1000],
            yrange=[0.0, 2.0],
            xlabel=r"$T_{\rm proton}$/MeV",
            ylabel=r"$E_{\ell^+} + E_{\ell^-}$/GeV",
            title=title + " proton-elastic only",
            nbins=n2D,
        )

        histogram2D(
            PATH / "2D_thetaLead_thetaSubLead.pdf",
            df["theta_subleading"],
            df["theta_leading"],
            w,
            xrange=[0.0, 40.0],
            yrange=[0.0, 40.0],
            xlabel=r"$\theta_{\ell_{\rm sublead}}$ degrees",
            ylabel=r"$\theta_{\ell_{\rm lead}}$ degrees",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_thetaSubLead_dtheta.pdf",
            df["theta_subleading"],
            df["Delta_theta"],
            w,
            xrange=[0.0, 40.0],
            yrange=[0.0, 40.0],
            xlabel=r"$\theta_{\nu_{\rm proj} \ell_{\rm sublead}}$ degrees",
            ylabel=r"$\Delta \theta$ degrees",
            **args_2d,
        )

        histogram2D(
            PATH / "2D_thetaLead_dtheta.pdf",
            df["theta_leading"],
            df["Delta_theta"],
            w,
            xrange=[0.0, 40.0],
            yrange=[0.0, 40.0],
            xlabel=r"$\theta_{\nu_{\rm proj} \ell_{\rm lead}}$ degrees",
            ylabel=r"$\Delta \theta$ degrees",
            **args_2d,
        )

        # HISTOGRAMS 1D ####################################################
        # momentum exchange
        histogram1D(PATH / "1D_E_lp.pdf", df["E_lp"], w, r"$E_{\ell^+}$/GeV", title, 100, TMIN=0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_E_lm.pdf", df["E_lm"], w, r"$E_{\ell^-}$/GeV", title, 100, TMIN=0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_E_tot.pdf", df["E_tot"], w, r"$(E_{\ell^-}+E_{\ell^+})$/GeV", title, 100, TMIN=0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_E_nu_QEreco.pdf", df["E_nu_reco"], w, r"$E_\nu^{\rm QE-reco}/$GeV", title, 20, TMIN=0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_E_leading.pdf", df["E_leading"], w, r"$E_{\rm leading}$/GeV", title, 100, TMIN=0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_E_subleading.pdf", df["E_subleading"], w, r"$E_{\rm subleading}$/GeV", title, 100, TMIN=0, TMAX=1.5, **args)
        histogram1D(PATH / "1D_cost_sum.pdf", df["costheta_sum"], w, r"$\cos(\theta_{(\ell \ell)\nu_{\rm proj}})$", title, 20, **args)
        histogram1D(PATH / "1D_theta_sum.pdf", df["theta_sum"], w, r"$\theta_{(\ell \ell) \nu_{\rm proj}}$", title, 20, TMIN=0, TMAX=60, **args)
        histogram1D(PATH / "1D_cost_sum_had.pdf", df["costheta_sum_had"], w, r"$\cos(\theta_{(\ell \ell) {\rm hadron}})$", title, 20, **args)
        histogram1D(PATH / "1D_theta_sum_had.pdf", df["theta_sum_had"], w, r"$\theta_{(\ell \ell) {\rm hadron}}$", title, 20, **args)
        histogram1D(PATH / "1D_cost_lp.pdf", df["costheta_lp"], w, r"$\cos(\theta_{\nu_{\rm proj} \ell^+})$", title, 40, **args)
        histogram1D(PATH / "1D_cost_lm.pdf", df["costheta_lm"], w, r"$\cos(\theta_{\nu_{\rm proj} \ell^-})$", title, 40, **args)
        histogram1D(PATH / "1D_theta_lp.pdf", df["theta_lp"], w, r"$\theta_{\nu_{\rm proj} \ell^+}$ degree", title, 40, **args)
        histogram1D(PATH / "1D_theta_lm.pdf", df["theta_lm"], w, r"$\theta_{\nu_{\rm proj} \ell^-}$ degree", title, 40, **args)
        histogram1D(PATH / "1D_theta_lead.pdf", df["theta_leading"], w, r"$\theta_{\nu_{\rm proj} \ell_{\rm lead}}$ degrees", title, 40, **args)
        histogram1D(PATH / "1D_theta_sublead.pdf", df["theta_subleading"], w, r"$\theta_{\nu_{\rm proj} \ell_{\rm sublead}}$ degrees", title, 40, **args)
        histogram1D(PATH / "1D_deltacos.pdf", df["Delta_costheta"], w, r"$\cos(\Delta \theta_{\ell^+ \ell^-})$", title, 40, **args)
        histogram1D(PATH / "1D_deltatheta.pdf", df["Delta_theta"], w, r"$\Delta \theta_{\ell^+ \ell^-}$ degrees", title, 40, TMIN=0, TMAX=60, **args)
        histogram1D(PATH / "1D_invmass.pdf", df["inv_mass"], w, r"$m_{\ell^+ \ell^-}$/GeV", title, 50, **args)
        histogram1D(PATH / "1D_asym.pdf", df["E_asy"], w, r"$(E_{\ell^+}-E_{\ell^-})$/($E_{\ell^+}+E_{\ell^-}$)", title, 20, TMIN=-1, TMAX=1, **args)
        histogram1D(PATH / "1D_asym_abs.pdf", np.abs(df["E_asy"]), w, r"$|E_{\ell^+}-E_{\ell^-}|$/($E_{\ell^+}+E_{\ell^-}$)", title, 20, TMIN=0, TMAX=1, **args)


def plot_closed_region(points, logx=False, logy=False):
    x, y = points
    if logy:
        if (y == 0).any():
            raise ValueError("y values cannot contain any zeros in log mode.")
        sy = np.sign(y)
        ssy = (np.abs(y) < 1) * (-1) + (np.abs(y) > 1) * (1)
        y = ssy * np.log(y * sy)
    if logx:
        if (x == 0).any():
            raise ValueError("x values cannot contain any zeros in log mode.")
        sx = np.sign(x)
        ssx = (x < 1) * (-1) + (x > 1) * (1)
        x = ssx * np.log(x * sx)

    points = np.array([x, y]).T

    points_s = points - points.mean(0)
    angles = np.angle((points_s[:, 0] + 1j * points_s[:, 1]))
    points_sort = points_s[angles.argsort()]
    points_sort += points.mean(0)

    tck, u = splprep(points_sort.T, u=None, s=0.0, per=0, k=1)
    u_new = np.linspace(u.min(), u.max(), len(points[:, 0]))
    x_new, y_new = splev(u_new, tck, der=0)

    if logx:
        x_new = sx * np.exp(ssx * x_new)
    if logy:
        y_new = sy * np.exp(ssy * y_new)

    return x_new, y_new



import numpy as np
import numpy.ma as ma

from DarkNews import logger

from . import Cfourvec as Cfv
from . import const
from . import fourvec


########################################################
# Kinematical limits

# 2 --> 2 scattering
def upscattering_Q2max(Enu, mHNL, M):
    s = 2 * Enu * M + M ** 2
    return (
        1
        / 2
        * (s) ** (-1)
        * (
            (M) ** (4)
            + (
                -1 * (mHNL) ** (2) * s
                + (
                    (s) ** (2)
                    + (
                        -1 * (M) ** (2) * ((mHNL) ** (2) + 2 * s)
                        + (-1 * (M) ** (2) + s) * (((M) ** (4) + ((((mHNL) ** (2) + -1 * s)) ** (2) + -2 * (M) ** (2) * ((mHNL) ** (2) + s)))) ** (1 / 2)
                    )
                )
            )
        )
    )


def upscattering_Q2min(Enu, mHNL, M):
    s = 2 * Enu * M + M ** 2
    r = mHNL / np.sqrt(s)
    m = M / np.sqrt(s)

    small_r = r < 1e-3
    # large cancellations at play -- expanding for small r
    q2min = ma.masked_array(
        data=1/2 * (1+ ((m) ** (4)+ (-1 * (r) ** (2)+ (-1 * ((((-1 + (m) ** (2))) ** (2) + (-2 * (1 + (m) ** (2)) * (r) ** (2) + (r) ** (4)))) ** (1 / 2)+ (m) ** (2) * (-2 + (-1 * (r) ** (2) + ((((-1 + (m) ** (2))) ** (2) + (-2 * (1 + (m) ** (2)) * (r) ** (2) + (r) ** (4)))) ** (1 / 2)))))))* s,
        mask=small_r,
        fill_value=(m) ** (2) * ((-1 + (m) ** (2))) ** (-2) * (r) ** (4) * s,
    )
    return q2min.filled()

    # q2min = np.piecewise(r, [r>=1e-3, r<1e-3], \
    # 						[lambda r: 1/2 * ( 1 + ( ( m )**( 4 ) + ( -1 * ( r )**( 2 ) + ( -1 * ( ( ( ( -1 + ( m )**( 2 ) ) )**( 2 ) + ( -2 * ( 1 + ( m )**( 2 ) ) * ( r )**( 2 ) + ( r )**( 4 ) ) ) )**( 1/2 ) + ( m )**( 2 ) * ( -2 + ( -1 * ( r )**( 2 ) + ( ( ( ( -1 + ( m )**( 2 ) ) )**( 2 ) + ( -2 * ( 1 + ( m )**( 2 ) ) * ( r )**( 2 ) + ( r )**( 4 ) ) ) )**( 1/2 ) ) ) ) ) ) ) * s,\
    # 						 lambda r: ( m )**( 2 ) * ( ( -1 + ( m )**( 2 ) ) )**( -2 ) * ( r )**( 4 ) * s])
    # return q2min


# 1 --> 3 decays (decay mandelstam)
def three_body_umax(m1, m2, m3, m4, t):
    return 1 / 4 * (((m1) ** (2) + ((m2) ** (2) + (-1 * (m3) ** (2) + -1 * (m4) ** (2))))) ** (2) * (t) ** (-1) + -1 * (
        (
            ((-1 * (m2) ** (2) + 1 / 4 * (t) ** (-1) * (((m2) ** (2) + (-1 * (m3) ** (2) + t))) ** (2))) ** (1 / 2)
            + -1 * ((-1 * (m4) ** (2) + 1 / 4 * (t) ** (-1) * ((-1 * (m1) ** (2) + ((m4) ** (2) + t))) ** (2))) ** (1 / 2)
        )) ** (2)


def three_body_umin(m1, m2, m3, m4, t):
    return 1 / 4 * (((m1) ** (2) + ((m2) ** (2) + (-1 * (m3) ** (2) + -1 * (m4) ** (2))))) ** (2) * (t) ** (-1) + -1 * (
        (
            ((-1 * (m2) ** (2) + 1 / 4 * (t) ** (-1) * (((m2) ** (2) + (-1 * (m3) ** (2) + t))) ** (2))) ** (1 / 2)
            + ((-1 * (m4) ** (2) + 1 / 4 * (t) ** (-1) * ((-1 * (m1) ** (2) + ((m4) ** (2) + t))) ** (2))) ** (1 / 2)
        )) ** (2)


def three_body_tmax(m1, m2, m3, m4):
    return (m1 - m4) ** 2


def three_body_tmin(m1, m2, m3, m4):
    return (m2 + m3) ** 2


# ######################
# 2 -> 2 scattering
# Ni(k1) target(k2) -->  Nj(k3) target(k4)
def two_to_two_scatter(samples, m1=1.0, m2=0.0, m3=1.0, m4=0.0):

    if "Eprojectile" in samples.keys():
        Eprojectile = samples["Eprojectile"]
        s = m2 ** 2 + 2 * Eprojectile * m2 + m1 ** 2
        sample_size = np.shape(Eprojectile)[0]
    else:
        logger.error("Error! Could not determine the projectile energy")

    E1CM = (s + m1 ** 2 - m2 ** 2) / 2.0 / np.sqrt(s)
    E2CM = (s - m1 ** 2 + m2 ** 2) / 2.0 / np.sqrt(s)
    E3CM = (s + m3 ** 2 - m4 ** 2) / 2.0 / np.sqrt(s)
    E4CM = (s - m3 ** 2 + m4 ** 2) / 2.0 / np.sqrt(s)

    p1CM = np.sqrt(E1CM ** 2 - m1 ** 2)
    p2CM = np.sqrt(E2CM ** 2 - m2 ** 2)
    p3CM = np.sqrt(E3CM ** 2 - m3 ** 2)
    p4CM = np.sqrt(E4CM ** 2 - m4 ** 2)

    Q2min = -(m1 ** 2 + m3 ** 2 - 2 * (E1CM * E3CM - p1CM * p3CM))
    Q2max = -(m1 ** 2 + m3 ** 2 - 2 * (E1CM * E3CM + p1CM * p3CM))

    if "unit_Q2" in samples.keys():
        Q2l = (np.log(Q2max) - np.log(Q2min)) * samples["unit_Q2"] + np.log(Q2min)
        Q2 = np.exp(Q2l)
    elif "Q2" in samples.keys():
        Q2 = samples["Q2"]
    else:
        logger.debug("DEBUG: Could not find Q2 samples, using uniform distribution instead.")
        Q2 = Cfv.random_generator(sample_size, Q2min, Q2max)

    # KINEMATICS TO LAB FRAME
    costN = (-Q2 - m1 ** 2 - m3 ** 2 + 2 * E1CM * E3CM) / (2 * p1CM * p3CM)
    beta = -p2CM / E2CM  # MINUS SIGN -- from CM to LAB
    gamma = 1.0 / np.sqrt(1.0 - beta * beta)

    if "unit_phi3" in samples.keys():
        phi3 = 2 * np.pi * samples["unit_phi3"]
    elif "phi3" in samples.keys():
        phi3 = samples["phi3"]
    else:
        phi3 = Cfv.random_generator(sample_size, 0.0, 2 * np.pi)

    P1CM = Cfv.build_fourvec(E1CM, p1CM, np.full_like(costN, 1.0), np.full_like(phi3, 0))
    P2CM = Cfv.build_fourvec(E2CM, -p1CM, np.full_like(costN, 1.0), np.full_like(phi3, 0))
    P3CM = Cfv.build_fourvec(E3CM, p3CM, costN, phi3)
    P4CM = Cfv.build_fourvec(E4CM, -p3CM, costN, phi3)

    # incoming neutrino
    P1LAB = Cfv.L(P1CM, beta)
    # incoming Hadron
    P2LAB = Cfv.L(P2CM, beta)
    # outgoing neutrino
    P3LAB = Cfv.L(P3CM, beta)
    # outgoing Hadron
    P4LAB = Cfv.L(P4CM, beta)

    return P1LAB, P2LAB, P3LAB, P4LAB


# Two body decay
# p1 (k1) --> p2(k2) p3(k3)
def two_body_decay(samples, boost=False, m1=1, m2=0, m3=0):

    if not samples:
        logger.debug("DEBUG: No samples were passed to two_body_decay. Assuming uniform phase space.")
        sample_size = np.shape(list(boost.values())[0])[0]
    else:
        # get sample size of the first item
        sample_size = np.shape(list(samples.values())[0])[0]

    # cosine of the angle between k3 and z axis
    if "unit_cost" in samples.keys():
        cost = 2 * samples["unit_cost"] - 1
    elif "cost" in samples.keys():
        cost = samples["cost"]
    else:
        logger.debug("DEBUG: Could not find cost samples, using uniform distribution instead.")
        cost = Cfv.random_generator(sample_size, -1, 1)

    E1CM_decay = np.full_like(cost, m1)
    E2CM_decay = np.full_like(cost, (m1 ** 2 + m2 ** 2 - m3 ** 2) / 2.0 / m1)
    E3CM_decay = np.full_like(cost, (m1 ** 2 - m2 ** 2 + m3 ** 2) / 2.0 / m1)

    p2CM_decay = np.full_like(cost, np.sqrt(E2CM_decay ** 2 - m2 ** 2))
    p3CM_decay = np.full_like(cost, np.sqrt(E3CM_decay ** 2 - m3 ** 2))

    # azimuthal angle of k3
    if "unit_phiz" in samples.keys():
        phiz = 2 * samples["unit_phiz"] - 1
    elif "phiz" in samples.keys():
        phiz = samples["phiz"]
    else:
        logger.debug("DEBUG: Could not find phiz samples, using uniform distribution instead.")
        phiz = Cfv.random_generator(sample_size, 0.0, 2 * np.pi)

    P1CM_decay = Cfv.build_fourvec(E1CM_decay, p2CM_decay * 0.0, cost / cost, phiz * 0)
    P2CM_decay = Cfv.build_fourvec(E2CM_decay, -p2CM_decay, cost, phiz)
    P3CM_decay = Cfv.build_fourvec(E3CM_decay, p2CM_decay, cost, phiz)

    # four-momenta in the LAB frame
    if boost:
        EP_LAB = boost["EP_LAB"]
        p1_LAB = np.sqrt(EP_LAB ** 2 - m1 ** 2)
        costP_LAB = boost["costP_LAB"]
        phiP_LAB = boost["phiP_LAB"]

        ### Transform P2_CM into the LAB frame (determined by the PN vector)

        P1LAB_decay = Cfv.build_fourvec(EP_LAB, p1_LAB, costP_LAB, phiP_LAB)

        P2LAB_decay = Cfv.Tinv(P2CM_decay, -p1_LAB / EP_LAB, costP_LAB, phiP_LAB)
        P3LAB_decay = Cfv.Tinv(P3CM_decay, -p1_LAB / EP_LAB, costP_LAB, phiP_LAB)

        return P1LAB_decay, P2LAB_decay, P3LAB_decay

    else:
        return P1CM_decay, P2CM_decay, P3CM_decay


# Three body decay
# p1 (k1) --> p2(k2) p3(k3) p4(k4)
def three_body_decay(samples, boost=False, m1=1, m2=0, m3=0, m4=0):

    if not samples:
        logger.error("Error! No samples were passed to three_body_decay.")
    else:
        # get sample size of the first item
        sample_size = np.shape(list(samples.values())[0])[0]

    # Mandelstam t = m23^2
    tminus = (m2 + m3) ** 2
    tplus = (m1 - m4) ** 2
    if "unit_t" in samples.keys():
        t = (tplus - tminus) * samples["unit_t"] + tminus
    elif "t" in samples.keys():
        t = samples["t"]
    else:
        logger.debug("DEBUG: Could not find t samples, using uniform distribution instead.")
        t = Cfv.random_generator(sample_size, tminus, tplus)

    # Mandelstam u = m_24^2
    # from MATHEMATICA
    uplus = 1 / 4 * (((m1) ** (2) + ((m2) ** (2) + (-1 * (m3) ** (2) + -1 * (m4) ** (2))))) ** (2) * (t) ** (-1) + -1 * (
        (
            ((-1 * (m2) ** (2) + 1 / 4 * (t) ** (-1) * (((m2) ** (2) + (-1 * (m3) ** (2) + t))) ** (2))) ** (1 / 2)
            + -1 * ((-1 * (m4) ** (2) + 1 / 4 * (t) ** (-1) * ((-1 * (m1) ** (2) + ((m4) ** (2) + t))) ** (2))) ** (1 / 2)
        )
    ) ** (2)
    # from MATHEMATICA
    uminus = 1 / 4 * (((m1) ** (2) + ((m2) ** (2) + (-1 * (m3) ** (2) + -1 * (m4) ** (2))))) ** (2) * (t) ** (-1) + -1 * (
        (
            ((-1 * (m2) ** (2) + 1 / 4 * (t) ** (-1) * (((m2) ** (2) + (-1 * (m3) ** (2) + t))) ** (2))) ** (1 / 2)
            + ((-1 * (m4) ** (2) + 1 / 4 * (t) ** (-1) * ((-1 * (m1) ** (2) + ((m4) ** (2) + t))) ** (2))) ** (1 / 2)
        )
    ) ** (2)
    if "unit_u" in samples.keys():
        u = (uplus - uminus) * samples["unit_u"] + uminus
    elif "u" in samples.keys():
        u = samples["u"]
    else:
        logger.debug("DEBUG: Could not find u samples, using uniform distribution instead.")
        u = Cfv.random_generator(sample_size, uminus, uplus)

    # Mandelstam v = m_34^2
    v = m1 ** 2 + m2 ** 2 + m3 ** 2 + m4 ** 2 - u - t

    E2CM_decay = (m1 ** 2 + m2 ** 2 - v) / 2.0 / m1
    E3CM_decay = (m1 ** 2 + m3 ** 2 - u) / 2.0 / m1
    E4CM_decay = (m1 ** 2 + m4 ** 2 - t) / 2.0 / m1

    p2CM_decay = np.sqrt(E2CM_decay * E2CM_decay - m2 ** 2)
    p3CM_decay = np.sqrt(E3CM_decay * E3CM_decay - m3 ** 2)
    p4CM_decay = np.sqrt(E4CM_decay * E4CM_decay - m4 ** 2)

    # Polar angle of P_3
    if "unit_c3" in samples.keys():
        c_theta3 = 2 * samples["unit_c3"] - 1
    elif "c3" in samples.keys():
        c_theta3 = samples["c3"]
    else:
        logger.debug("DEBUG: Could not find c3 samples, using uniform distribution instead.")
        c_theta3 = Cfv.random_generator(sample_size, -1, 1)

    phi3 = Cfv.random_generator(sample_size, 0.0, 2 * np.pi)

    # Azimuthal angle of P_4 wrt to P_3 (phi_34)
    if "unit_phi34" in samples.keys():
        phi34 = 2 * np.pi * samples["unit_phi34"]
    elif "phi34" in samples.keys():
        phi34 = samples["phi34"]
    else:
        logger.debug("DEBUG: Could not find phi34 samples, using uniform distribution instead.")
        phi34 = Cfv.random_generator(sample_size, 0, 2 * np.pi)

    # polar angle of P_4 wrt to P_3 is a known function of u and v
    c_theta34 = (t + u - m2 ** 2 - m1 ** 2 + 2 * E3CM_decay * E4CM_decay) / (2 * p3CM_decay * p4CM_decay)

    # p1
    P1CM_decay = Cfv.build_fourvec(m1 * np.ones(sample_size), np.zeros(sample_size), np.ones(sample_size), np.zeros(sample_size))
    # p3
    P3CM_decay = Cfv.build_fourvec(E3CM_decay, p3CM_decay, c_theta3, phi3)
    # p4 -- first in p3 along Z, then rotated to CM frame
    P4CM_decay = Cfv.rotationz(Cfv.rotationy_cos(Cfv.build_fourvec(E4CM_decay, p4CM_decay, c_theta34, phi34), c_theta3, sign=-1), phi3)
    # p2
    P2CM_decay = P1CM_decay - P4CM_decay - P3CM_decay

    # four-momenta in the LAB frame
    if boost:
        EN_LAB = boost["EP_LAB"]
        costN_LAB = boost["costP_LAB"]
        phiN_LAB = boost["phiP_LAB"]

        ### Transform from CM into the LAB frame
        # Decaying neutrino
        P1LAB_decay = Cfv.Tinv(P1CM_decay, -np.sqrt(EN_LAB ** 2 - m1 ** 2) / EN_LAB, costN_LAB, phiN_LAB)
        # Outgoing neutrino
        P2LAB_decay = Cfv.Tinv(P2CM_decay, -np.sqrt(EN_LAB ** 2 - m1 ** 2) / EN_LAB, costN_LAB, phiN_LAB)
        # Outgoing lepton minus (3)
        P3LAB_decay = Cfv.Tinv(P3CM_decay, -np.sqrt(EN_LAB ** 2 - m1 ** 2) / EN_LAB, costN_LAB, phiN_LAB)
        # Outgoing lepton plus (4)
        P4LAB_decay = Cfv.Tinv(P4CM_decay, -np.sqrt(EN_LAB ** 2 - m1 ** 2) / EN_LAB, costN_LAB, phiN_LAB)

        return P1LAB_decay, P2LAB_decay, P3LAB_decay, P4LAB_decay

    # four-momenta in the parent particle rest frame
    else:
        return P1CM_decay, P2CM_decay, P3CM_decay, P4CM_decay



import numpy as np
from . import const
from . import Cfourvec as Cfv


class Chisel:
    def __init__(self, nsamples, box=np.array(3 * [[-0.5, 0.5]]), name="my_mold"):
        """
        Sculptor of detector geometries. We start with a cuboid full of uniformly generated points and
        start sculpting the desired detector geometry by iteratively cutting events in given 3D objects.
        One can then find the union or intersection of these objects. The available objects are:
                * rectangle
                * sphere
                * cylinder
                * spherical_cap

        More complex and final pre-defined objects are also available:
                * the microboone cryostat as a junction of a tubbe with two spherical caps.

        A desired set of points randomly distributed across a final 3D object can then found by iteratively
        creating more sculptors with the same Chisel, until the desired number of points is reached.

        Args:
                nsamples (int): number of locations to generate
                box (numpy.ndarray, optional): size of the initial cuboid. Defaults to a cube: np.array(3*[[-0.5,0.5]]).
                name (str, optional): _description_. Defaults to 'my_mold'.

        Raises:
                ValueError: raised when the chisel suspects that the detector geometry is not specified correctly.
        """

        self.name = name
        self.nsamples = nsamples

        # initialize the ivory box
        if (box[:, 1] - box[:, 0] < 0).any():
            raise ValueError(f"Box axis reversed, x_0 > x_1 for dimensions: {[i for i, x in enumerate(box[:,1] - box[:,0]) if x < 0 ]}.")
        self.x = (box[0, 1] - box[0, 0]) * np.random.rand(nsamples) + box[0, 0]
        self.y = (box[1, 1] - box[1, 0]) * np.random.rand(nsamples) + box[1, 0]
        self.z = (box[2, 1] - box[2, 0]) * np.random.rand(nsamples) + box[2, 0]

        self.events = np.array([self.x, self.y, self.z])

    def union(self, mold1, mold2):
        return mold1 | mold2

    def intersection(self, mold1, mold2):
        return mold1 & mold2

    def translate(self, xvec, translation):
        return xvec + translation

    # simple geometrical objects
    def rectangle(self, dx, dy, dz, origin=np.zeros((3, 1))):
        x, y, z = self.translate(origin, self.events)
        return (-dx / 2 < x & x < dx / 2) & (-dy / 2 < y & y < dy / 2) & (-dz / 2 < z & z < dz / 2)

    def sphere(self, radius, origin=np.zeros((3, 1))):
        x, y, z = self.translate(origin, self.events)
        return x ** 2 + y ** 2 + z ** 2 < radius ** 2

    def cylinder(self, radius, height, origin=np.zeros((3, 1))):
        x, y, z = self.translate(origin, self.events)
        return (x ** 2 + y ** 2 < radius ** 2) & (-height / 2 < z) & (z < height / 2)

    def spherical_cap(self, radius, zenith_max, origin=np.zeros((3, 1))):
        x, y, z = self.translate(origin, self.events)
        rpolar = np.sqrt(x ** 2 + y ** 2)
        r = np.sqrt(rpolar ** 2 + z ** 2)
        zenith = np.arctan2(z, rpolar)
        h = radius * (1 - np.cos(zenith_max))
        return (r < radius) | (zenith < zenith_max) | (z < -radius + h)

    # def microboone_cryostat(self):

    # 	# Tube parameters
    # 	r_t = 191.61
    # 	z_t  = 1086.49

    # 	# cap parameters
    # 	r_c = 305.250694958
    # 	theta_c = 38.8816337686*const.deg_to_rad
    # 	ctheta_c = np.cos(theta_c)

    # 	h = r_c*(1.-ctheta_c)
    # 	cap_gap = 0.37361008
    # 	z_cap = - z_t/2 - cap_gap + r_c*ctheta_c

    # 	tube =	self.cylinder(radius=r_t, height=z_t)
    # 	cap1 =	self.spherical_cap(radius=r_c, zenith_max=theta_c, origin=np.array([[0],[0], [-z_cap]]))
    # 	cap2 =	self.spherical_cap(radius=r_c, zenith_max=theta_c, origin=np.array([[0],[0], [ z_cap]]))

    # 	return tube | cap1 | cap2

    # all in centimeters
    def microboone_cryostat(self):

        # Tube parameters
        r_t = 191.61
        z_t = 1086.49

        # cap parameters
        r_c = 305.250694958
        theta_c = 38.8816337686 * const.deg_to_rad
        ctheta_c = np.cos(theta_c)
        zend_c = 305.624305042
        h = r_c * (1.0 - ctheta_c)
        cap_gap = 0.37361008

        x = self.x
        y = self.y
        z = self.z

        # accept in tube
        r_polar = np.sqrt(x ** 2 + y ** 2)
        mask_tube = (-z_t / 2 < z) & (z < z_t / 2) & (r_polar < r_t)

        # coordinates in sphere 1
        z1 = z - (-z_t / 2 - cap_gap + r_c * ctheta_c)
        r1 = np.sqrt(r_polar ** 2 + z1 ** 2)
        inc1 = np.arctan2(z1, r_polar)  # inclination angle

        # accept in front cap
        mask_cap1 = (r1 < r_c) & (inc1 < theta_c) & (z1 < -r_c + h)

        # coordinates in sphere 2
        z2 = z + (-z_t / 2 - cap_gap - h + r_c)
        r2 = np.sqrt(r_polar ** 2 + z2 ** 2)
        inc2 = np.arctan2(-z2, r_polar)  # inclination angle

        # accept in back cap
        mask_cap2 = (r2 < r_c) & (inc2 < theta_c) & (z2 > r_c - h)

        mask_full = mask_tube + mask_cap1 + mask_cap2

        return mask_full


# distribute events in df accross the pre-defined MicroBooNE (cryostat) volume
def microboone_geometry(df):

    nsamples = len(df.index)

    detector_box = np.array([[-200, 200], [-200, 200], [-800, 800]])

    tries = 0
    npoints = 0
    events = np.array(3 * [[]])
    while npoints < nsamples:

        new_detector = Chisel(nsamples=nsamples, box=detector_box)
        new_events = new_detector.events[:, new_detector.microboone_cryostat()]
        events = np.concatenate((events, new_events), axis=1)

        npoints += np.shape(new_events)[1]
        tries += nsamples
        if tries > 1e3 * nsamples:
            raise ValueError("Geometry sampled too inefficiently. Wrong setup?")

    # Parameters from Mark -- neutrino time spill, overlapping w Genie BNB
    GlobalTimeOffset = 3125.0
    RandomTimeOffset = 1600.0
    time = GlobalTimeOffset + (RandomTimeOffset) * np.random.rand(nsamples)

    # guarantee that array has number of samples asked (nsamples)
    df["pos_scatt", "0"] = time
    df["pos_scatt", "1"] = events[0, :nsamples]
    df["pos_scatt", "2"] = events[1, :nsamples]
    df["pos_scatt", "3"] = events[2, :nsamples]


# distribute events in df accross the pre-defined spherical MiniBooNE volume
def miniboone_geometry(df):

    nsamples = len(df.index)

    detector_box = np.array([[-600, 600], [-600, 600], [-600, 600]])

    tries = 0
    npoints = 0
    events = np.array(3 * [[]])
    while npoints < nsamples:

        new_detector = Chisel(nsamples=nsamples, box=detector_box)
        new_events = new_detector.events[:, new_detector.sphere(radius=574.6)]
        events = np.concatenate((events, new_events), axis=1)

        npoints += np.shape(new_events)[1]
        tries += nsamples
        if tries > 1e3 * nsamples:
            raise ValueError(f"Geometry sampled too inefficiently, tries = {tries} and npoints = {npoints}. Wrong setup?")

    # guarantee that array has number of samples asked (nsamples)
    df["pos_scatt", "0"] = np.zeros((nsamples,))
    df["pos_scatt", "1"] = events[0, :nsamples]
    df["pos_scatt", "2"] = events[1, :nsamples]
    df["pos_scatt", "3"] = events[2, :nsamples]


# assing all events in df a scattering position at 4-position (0,0,0,0)
def point_geometry(df):
    nsamples = len(df.index)
    df["pos_scatt", "0"] = np.zeros((nsamples,))
    df["pos_scatt", "1"] = np.zeros((nsamples,))
    df["pos_scatt", "2"] = np.zeros((nsamples,))
    df["pos_scatt", "3"] = np.zeros((nsamples,))


def place_decay(df, df_column, l_decay_proper_cm, label="decay_pos"):
    """find decay positions given scattering positions and lifetime (sample from exponential) and
            save that in the dataframe (in-place)

    Args:
            df (pd.DataFrame): _description_
            df_column (str): column of the dataframe to use -- this determines which particle we are using
            l_decay_proper_cm (float): the proper decay length of the particle (obtained from vegas or otherwise)
            label (str, optional): label to be used for the dataframe new columns. Defaults to 'decay_pos'.
    """
    # four momentum as numpy array
    p = df[df_column].to_numpy()

    # parent mass
    M = np.sqrt(Cfv.dot4(p, p))
    # momentum
    pvec = np.sqrt(p[:, 0] ** 2 - M ** 2)
    beta = pvec / p[:, 0]
    gamma = 1 / np.sqrt(1 - beta ** 2)
    gammabeta = gamma * beta

    ######################
    # sample displacement from decay propability
    d_decay = np.random.exponential(scale=l_decay_proper_cm * gammabeta)  # centimeters

    # direction of N
    df[label, "0"] = df["pos_scatt", "0"] + d_decay / (beta * const.c_LIGHT)
    df[label, "1"] = df["pos_scatt", "1"] + Cfv.get_3direction(p)[:, 0] * d_decay
    df[label, "2"] = df["pos_scatt", "2"] + Cfv.get_3direction(p)[:, 1] * d_decay
    df[label, "3"] = df["pos_scatt", "3"] + Cfv.get_3direction(p)[:, 2] * d_decay